from datetime import date, timedelta

from dateutil import relativedelta as rd

from ..core import WesternCalendar, ChristianMixin
from ..core import SUN, SAT
from ..core import Holiday
from ..registry_tools import iso_register


@iso_register('MX')
class Mexico(WesternCalendar, ChristianMixin):
    "Mexico"
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (5, 1, "Labour Day"),
        (9, 16, "Independence Day"),
    )

    shift_new_years_day = True

    @property
    def observance_shift(self):
        return Holiday.nearest_weekday

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(Holiday(
            date(year, 2, 1) + rd.relativedelta(weekday=rd.MO(1)),
            "Constitution Day",
        ))

        days.append(Holiday(
            date(year, 3, 1) + rd.relativedelta(weekday=rd.MO(3)),
            "Benito Juárez's birthday",
        ))

        days.append(Holiday(
            date(year, 11, 1) + rd.relativedelta(weekday=rd.MO(3)),
            "Revolution Day",
        ))

        return days

    def get_calendar_holidays(self, year):
        days = super().get_calendar_holidays(year)
        # If any statutory day is on Sunday, the monday is off
        # If it's on a Saturday, the Friday is off
        for day, label in days:
            if day.weekday() == SAT:
                days.append((day - timedelta(days=1), f"{label} substitute"))
            elif day.weekday() == SUN:
                days.append((day + timedelta(days=1), f"{label} substitute"))
        # Extra: if new year's day is a saturday, the friday before is off
        next_new_year = date(year + 1, 1, 1)
        if next_new_year.weekday():
            days.append((date(year, 12, 31), "New Year Day substitute"))
        return days



from ..core import OrthodoxCalendar
from ..registry_tools import iso_register


@iso_register('GR')
class Greece(OrthodoxCalendar):
    'Greece'

    # Civil holidays
    include_labour_day = True
    FIXED_HOLIDAYS = OrthodoxCalendar.FIXED_HOLIDAYS + (
        (3, 25, "Independence Day"),
        (10, 28, "Ohi Day"),
    )

    # Christian holidays
    include_epiphany = True
    include_clean_monday = True
    include_annunciation = True
    include_good_friday = True
    include_easter_sunday = True
    include_easter_monday = True
    include_whit_sunday = True
    whit_sunday_label = "Pentecost"
    include_whit_monday = True
    include_assumption = True
    include_boxing_day = True
    boxing_day_label = "Glorifying Mother of God"
    # No Orthodox Christmas
    include_orthodox_christmas = False



from .china import China
from .hong_kong import HongKong, HongKongBank
from .japan import Japan, JapanBank
from .malaysia import Malaysia
from .qatar import Qatar
from .singapore import Singapore
from .south_korea import SouthKorea
from .taiwan import Taiwan
from .israel import Israel


__all__ = (
    'China',
    'HongKong',
    'HongKongBank',
    'Japan',
    'JapanBank',
    'Malaysia',
    'Qatar',
    'Singapore',
    'SouthKorea',
    'Taiwan',
    'Israel',
)



from ..core import WesternCalendar
from ..registry_tools import iso_register


@iso_register('MG')
class Madagascar(WesternCalendar):
    "Madagascar"
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (3, 29, "Martyrs' Day"),
        (6, 26, "Independence Day"),
    )
    # Civil holidays
    include_labour_day = True
    # Christian holidays
    include_easter_monday = True
    include_ascension = True
    include_whit_monday = True
    include_assumption = True
    include_all_saints = True



"""
Spring Holiday mixins
"""
from datetime import timedelta

from ....core import MON


class SpringHoliday:
    include_spring_holiday = True


class SpringHolidayFirstMondayApril(SpringHoliday):
    def get_spring_holiday(self, year):
        return (
            self.get_nth_weekday_in_month(year, 4, MON),
            self.spring_holiday_label
        )


class SpringHolidaySecondMondayApril(SpringHoliday):
    def get_spring_holiday(self, year):
        return (
            self.get_nth_weekday_in_month(year, 4, MON, 2),
            self.spring_holiday_label
        )


class SpringHolidayTuesdayAfterFirstMondayMay(SpringHoliday):

    def get_spring_holiday(self, year):
        first_monday = self.get_nth_weekday_in_month(year, 5, MON)
        return (
            first_monday + timedelta(days=1),
            self.spring_holiday_label
        )


class SpringHolidayLastMondayMay(SpringHoliday):
    def get_spring_holiday(self, year):
        return (
            self.get_last_weekday_in_month(year, 5, MON),
            self.spring_holiday_label
        )


class SpringHolidayFirstMondayJune(SpringHoliday):
    def get_spring_holiday(self, year):
        return (
            self.get_nth_weekday_in_month(year, 6, MON),
            self.spring_holiday_label
        )



from datetime import date

from ..america import (Argentina, Barbados, Chile, Colombia, Mexico, Panama,
                       Paraguay)
from . import GenericCalendarTest


class ArgentinaTest(GenericCalendarTest):
    cal_class = Argentina

    def test_holidays_2018(self):
        holidays = self.cal.holidays_set(2018)
        # 1. Año Nuevo
        self.assertIn(date(2018, 1, 1), holidays)
        # 2. Carnaval
        self.assertIn(date(2018, 2, 12), holidays)
        # 3. Carnaval
        self.assertIn(date(2018, 2, 13), holidays)
        # 4. Día de la Memoria
        self.assertIn(date(2018, 3, 24), holidays)
        # 5. Día del Veterano y de los Caídos en la Guerra de Malvinas
        self.assertIn(date(2018, 4, 2), holidays)
        # 6. Viernes Santo
        self.assertIn(date(2018, 3, 30), holidays)
        # 7. Día del Trabajador
        self.assertIn(date(2018, 5, 1), holidays)
        # 8. Día de la Revolución de Mayo
        self.assertIn(date(2018, 5, 25), holidays)
        # 9. Día Paso a la Inmortalidad del General Manuel Belgrano
        self.assertIn(date(2018, 6, 20), holidays)
        # 10. Día de la Independencia
        self.assertIn(date(2018, 7, 9), holidays)
        # 11. Inmaculada Concepción de María
        self.assertIn(date(2018, 12, 8), holidays)
        # 12. Navidad
        self.assertIn(date(2018, 12, 25), holidays)
        # variable days
        # 13. Día Paso a la Inmortalidad del General Martín Miguel de Güemes
        self.assertIn(date(2018, 6, 17), holidays)
        # 14. Paso a la Inmortalidad del General José de San Martín
        self.assertIn(date(2018, 8, 20), holidays)
        # 15. Día del Respeto a la Diversidad Cultural
        self.assertIn(date(2018, 10, 15), holidays)
        # 16. Día de la Soberanía Nacional
        self.assertIn(date(2018, 11, 19), holidays)

    def test_holidays_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 1, 1), holidays)
        self.assertIn(date(2019, 3, 4), holidays)
        self.assertIn(date(2019, 3, 5), holidays)
        self.assertIn(date(2019, 3, 24), holidays)
        self.assertIn(date(2019, 4, 2), holidays)
        self.assertIn(date(2019, 4, 19), holidays)
        self.assertIn(date(2019, 5, 1), holidays)
        self.assertIn(date(2019, 5, 25), holidays)
        self.assertIn(date(2019, 6, 20), holidays)
        self.assertIn(date(2019, 7, 9), holidays)
        self.assertIn(date(2019, 12, 8), holidays)
        self.assertIn(date(2019, 12, 25), holidays)
        # variable days
        self.assertIn(date(2019, 6, 17), holidays)
        self.assertIn(date(2019, 8, 19), holidays)
        self.assertIn(date(2019, 10, 14), holidays)
        self.assertIn(date(2019, 11, 18), holidays)

    def test_holidays_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 1, 1), holidays)
        self.assertIn(date(2020, 2, 24), holidays)
        self.assertIn(date(2020, 2, 25), holidays)
        self.assertIn(date(2020, 3, 24), holidays)
        # Special case: Argentina has shifted this holiday due to
        # Coronavirus lockdown in 2020.
        self.assertNotIn(date(2020, 4, 2), holidays)
        self.assertIn(date(2020, 3, 31), holidays)

        # Back to normal, I hope...
        self.assertIn(date(2020, 4, 10), holidays)
        self.assertIn(date(2020, 5, 1), holidays)
        self.assertIn(date(2020, 5, 25), holidays)
        self.assertIn(date(2020, 6, 20), holidays)
        self.assertIn(date(2020, 7, 9), holidays)
        self.assertIn(date(2020, 12, 8), holidays)
        self.assertIn(date(2020, 12, 25), holidays)
        # variable days
        self.assertIn(date(2020, 6, 15), holidays)
        self.assertIn(date(2020, 8, 17), holidays)
        self.assertIn(date(2020, 10, 12), holidays)
        self.assertIn(date(2020, 11, 23), holidays)

    def test_holidays_2021(self):
        # Testing it because June 17th happens on THU (general_guemes_day).
        holidays = self.cal.holidays_set(2021)
        # Not happening on June 17
        self.assertNotIn(date(2021, 6, 17), holidays)
        # Happens on the 1st MON after this date.
        self.assertIn(date(2021, 6, 20), holidays)

        # Also, Día del Respeto a la Diversidad Cultural is shifted
        self.assertNotIn(date(2021, 10, 12), holidays)
        # The day before
        self.assertIn(date(2021, 10, 11), holidays)

    def test_dia_malvinas_label(self):
        _, label = self.cal.get_malvinas_day(2020)
        self.assertEqual(
            label,
            "Día del Veterano y de los Caídos en la Guerra de Malvinas"
        )

    def test_dia_memoria_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        label_memoria = holidays[date(2020, 3, 24)]
        self.assertEqual(
            label_memoria,
            "Día Nacional de la Memoria por la Verdad y la Justicia"
        )

    def test_carnival_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        label_carnival = holidays[date(2020, 2, 25)]
        self.assertEqual(label_carnival, "Carnival")

    def test_labour_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        label = holidays[date(2020, 5, 1)]
        self.assertEqual(label, "Día del Trabajador")

    def test_immaculate_conception_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        label = holidays[date(2020, 12, 8)]
        self.assertEqual(label, "Día de la Inmaculada Concepción de María")


class ChileTest(GenericCalendarTest):
    cal_class = Chile

    def test_holidays_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)
        self.assertIn(date(2013, 3, 29), holidays)
        self.assertIn(date(2013, 3, 30), holidays)
        self.assertIn(date(2013, 5, 1), holidays)
        self.assertIn(date(2013, 5, 21), holidays)
        self.assertIn(date(2013, 6, 29), holidays)
        self.assertIn(date(2013, 7, 16), holidays)
        self.assertIn(date(2013, 8, 15), holidays)
        self.assertIn(date(2013, 9, 18), holidays)
        self.assertIn(date(2013, 9, 19), holidays)
        self.assertIn(date(2013, 9, 20), holidays)
        self.assertIn(date(2013, 10, 12), holidays)
        self.assertIn(date(2013, 10, 31), holidays)
        self.assertIn(date(2013, 11, 1), holidays)
        self.assertIn(date(2013, 12, 8), holidays)
        self.assertIn(date(2013, 12, 25), holidays)
        self.assertIn(date(2013, 12, 31), holidays)

    def test_holidays_2021(self):
        holidays = self.cal.holidays_set(2021)
        # Año Nuevo
        self.assertIn(date(2021, 1, 1), holidays)
        # Viernes Santo
        self.assertIn(date(2021, 4, 2), holidays)
        # Sábado Santo
        self.assertIn(date(2021, 4, 3), holidays)
        # Día Nacional del Trabajo
        self.assertIn(date(2021, 5, 1), holidays)
        # Día de las Glorias Navales
        self.assertIn(date(2021, 5, 21), holidays)
        # día nacional de los pueblos indígenas (June solstice)
        self.assertIn(date(2021, 6, 21), holidays)
        # San Pedro y San Pablo
        self.assertIn(date(2021, 6, 28), holidays)
        # Día de la Virgen del Carmen
        self.assertIn(date(2021, 7, 16), holidays)
        # Asunción de la Virgen
        self.assertIn(date(2021, 8, 15), holidays)
        # Additional Holiday
        self.assertIn(date(2021, 9, 17), holidays)
        # Independencia Nacional
        self.assertIn(date(2021, 9, 18), holidays)
        # Día de las Glorias del Ejército
        self.assertIn(date(2021, 9, 19), holidays)
        # Encuentro de Dos Mundos
        self.assertIn(date(2021, 10, 11), holidays)
        # Día de las Iglesias Evangélicas y Protestantes
        self.assertIn(date(2021, 10, 31), holidays)
        # Día de Todos los Santos
        self.assertIn(date(2021, 11, 1), holidays)
        # Inmaculada Concepción
        self.assertIn(date(2021, 12, 8), holidays)
        # Navidad
        self.assertIn(date(2021, 12, 25), holidays)
        # Feriado Bancario
        self.assertIn(date(2021, 12, 31), holidays)

    def test_indigenous_people_day(self):
        # Testing because variable nature of June solstice

        # approved in 2021
        holidays = self.cal.holidays_set(2020)
        self.assertNotIn(date(2020, 6, 21), holidays)
        self.assertNotIn(date(2020, 6, 20), holidays)

        # fixed day in 2021
        holidays = self.cal.holidays_set(2021)
        self.assertIn(date(2021, 6, 21), holidays)

        # solstice 2022
        holidays = self.cal.holidays_set(2022)
        self.assertIn(date(2022, 6, 21), holidays)

        # solstice 2023
        holidays = self.cal.holidays_set(2023)
        self.assertIn(date(2023, 6, 21), holidays)

        # solstice 2023
        holidays = self.cal.holidays_set(2024)
        self.assertIn(date(2024, 6, 20), holidays)

    def test_reformation_day(self):
        holidays = self.cal.holidays_set(2012)
        self.assertNotIn(date(2012, 10, 31), holidays)
        self.assertIn(date(2012, 11, 2), holidays)

        holidays = self.cal.holidays_set(2017)
        self.assertNotIn(date(2017, 10, 31), holidays)
        self.assertIn(date(2017, 10, 27), holidays)

    def test_national_bridge_days(self):
        # MON TUE
        holidays = self.cal.holidays_set(2017)
        self.assertIn(date(2017, 9, 18), holidays)
        self.assertIn(date(2017, 9, 19), holidays)

        # TUE WED
        holidays = self.cal.holidays_set(2018)
        self.assertIn(date(2018, 9, 17), holidays)
        self.assertIn(date(2018, 9, 18), holidays)
        self.assertIn(date(2018, 9, 19), holidays)
        self.assertNotIn(date(2018, 9, 20), holidays)

        # WED THU
        holidays = self.cal.holidays_set(2019)
        self.assertNotIn(date(2019, 9, 17), holidays)
        self.assertIn(date(2019, 9, 18), holidays)
        self.assertIn(date(2019, 9, 19), holidays)
        self.assertIn(date(2019, 9, 20), holidays)

        # THU FRI
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 9, 18), holidays)
        self.assertIn(date(2020, 9, 19), holidays)

        # FRI SAT
        holidays = self.cal.holidays_set(2015)
        self.assertNotIn(date(2015, 9, 17), holidays)
        self.assertIn(date(2015, 9, 18), holidays)
        self.assertIn(date(2015, 9, 19), holidays)

        # SUN MON
        holidays = self.cal.holidays_set(2016)
        self.assertIn(date(2016, 9, 18), holidays)
        self.assertIn(date(2016, 9, 19), holidays)
        self.assertNotIn(date(2016, 9, 20), holidays)

        # SAT SUN for additional day
        holidays = self.cal.holidays_set(2021)
        self.assertIn(date(2021, 9, 17), holidays)
        self.assertIn(date(2021, 9, 18), holidays)
        self.assertIn(date(2021, 9, 19), holidays)

    def test_columbus_day(self):
        # MON
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 10, 12), holidays)
        # TUE
        holidays = self.cal.holidays_set(2021)
        self.assertNotIn(date(2021, 10, 12), holidays)
        self.assertIn(date(2021, 10, 11), holidays)
        # WED
        holidays = self.cal.holidays_set(2016)
        self.assertNotIn(date(2016, 10, 12), holidays)
        self.assertIn(date(2016, 10, 10), holidays)
        # THU
        holidays = self.cal.holidays_set(2017)
        self.assertNotIn(date(2017, 10, 12), holidays)
        self.assertIn(date(2017, 10, 9), holidays)
        # FRI
        holidays = self.cal.holidays_set(2018)
        self.assertNotIn(date(2018, 10, 12), holidays)
        self.assertIn(date(2018, 10, 15), holidays)

    def test_st_peter_and_st_paul_day(self):
        # MON
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 6, 29), holidays)
        # TUE
        holidays = self.cal.holidays_set(2021)
        self.assertNotIn(date(2021, 6, 29), holidays)
        self.assertIn(date(2021, 6, 28), holidays)
        # WED
        holidays = self.cal.holidays_set(2016)
        self.assertNotIn(date(2016, 6, 29), holidays)
        self.assertIn(date(2016, 6, 27), holidays)
        # THU
        holidays = self.cal.holidays_set(2017)
        self.assertNotIn(date(2017, 6, 29), holidays)
        self.assertIn(date(2017, 6, 26), holidays)
        # FRI
        holidays = self.cal.holidays_set(2018)
        self.assertNotIn(date(2018, 6, 29), holidays)
        self.assertIn(date(2018, 7, 2), holidays)


class ColombiaTest(GenericCalendarTest):
    cal_class = Colombia

    def test_holidays_2015(self):
        holidays = self.cal.holidays_set(2015)
        self.assertIn(date(2015, 1, 1), holidays)    # New year
        self.assertIn(date(2015, 1, 12), holidays)   # Epiphany (shifted)
        self.assertIn(date(2015, 3, 23), holidays)   # Saint Joseph
        self.assertIn(date(2015, 3, 29), holidays)   # Palm Sunday
        self.assertIn(date(2015, 4, 2), holidays)    # Holy Thursday
        self.assertIn(date(2015, 4, 3), holidays)    # Good Friday
        self.assertIn(date(2015, 4, 5), holidays)    # Easter (SUN)
        self.assertIn(date(2015, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2015, 5, 18), holidays)   # Ascension (shifted)
        self.assertIn(date(2015, 6, 8), holidays)    # Corpus Christi
        self.assertIn(date(2015, 6, 15), holidays)   # Sacred Heart
        self.assertIn(date(2015, 6, 29), holidays)   # St Peter & St Paul
        self.assertIn(date(2015, 7, 20), holidays)   # Independance Day
        self.assertIn(date(2015, 8, 7), holidays)    # Boyacá battle
        self.assertIn(date(2015, 8, 17), holidays)   # Assumption (shifted)
        self.assertIn(date(2015, 10, 12), holidays)  # Day of the Races
        self.assertIn(date(2015, 11, 2), holidays)   # All Saints (shifted)
        self.assertIn(date(2015, 11, 16), holidays)  # Cartagena independence
        self.assertIn(date(2015, 12, 8), holidays)   # Immaculate Conception
        self.assertIn(date(2015, 12, 25), holidays)  # XMas
        self.assertEqual(len(holidays), 20)

    def test_holidays_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 1, 1), holidays)    # New year
        self.assertIn(date(2020, 1, 6), holidays)    # Epiphany
        self.assertIn(date(2020, 3, 23), holidays)   # Saint Joseph
        self.assertIn(date(2020, 4, 5), holidays)    # Palm Sunday
        self.assertIn(date(2020, 4, 9), holidays)    # Holy Thursday
        self.assertIn(date(2020, 4, 10), holidays)   # Good Friday
        self.assertIn(date(2020, 4, 12), holidays)   # Easter (SUN)
        self.assertIn(date(2020, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2020, 5, 25), holidays)   # Ascension (shifted)
        self.assertIn(date(2020, 6, 15), holidays)   # Corpus Christi
        self.assertIn(date(2020, 6, 22), holidays)   # Sacred Heart
        self.assertIn(date(2020, 6, 29), holidays)   # St Peter & St Paul
        self.assertIn(date(2020, 7, 20), holidays)   # Independance Day
        self.assertIn(date(2020, 8, 7), holidays)    # Boyacá battle
        self.assertIn(date(2020, 8, 17), holidays)   # Assumption (shifted)
        self.assertIn(date(2020, 10, 12), holidays)  # Day of the Races
        self.assertIn(date(2020, 11, 2), holidays)   # All Saints (shifted)
        self.assertIn(date(2020, 11, 16), holidays)  # Cartagena independence
        self.assertIn(date(2020, 12, 8), holidays)   # Immaculate Conception
        self.assertIn(date(2020, 12, 25), holidays)  # XMas
        self.assertEqual(len(holidays), 20)

    def test_epiphany_monday(self):
        # In 2020, Epiphany falls on MON
        epiphany_2020 = self.cal.get_epiphany(2020)
        self.assertEqual(epiphany_2020, date(2020, 1, 6))
        # In 2021, it does not, so it's shifted to the next MON
        epiphany_2021 = self.cal.get_epiphany(2021)
        self.assertEqual(epiphany_2021, date(2021, 1, 11))

    def test_saint_peter_and_saint_paul_monday(self):
        # In 2020, Saint Peter and Saint Paul falls on MON
        st_peter_paul_2020 = self.cal.get_saint_peter_and_saint_paul(2020)
        self.assertEqual(st_peter_paul_2020, date(2020, 6, 29))
        # In 2021, it does not, so it's shifted to the next MON
        st_peter_paul_2021 = self.cal.get_saint_peter_and_saint_paul(2021)
        self.assertEqual(st_peter_paul_2021, date(2021, 7, 5))

    def test_assumption_monday(self):
        # In 2021, Assumption falls on SUN, so it's shifted to MON
        assumption_2021 = self.cal.get_assumption(2021)
        self.assertEqual(assumption_2021, date(2021, 8, 16))
        # In 2022, Assumption falls on MON
        assumption_2022 = self.cal.get_assumption(2022)
        self.assertEqual(assumption_2022, date(2022, 8, 15))

    def test_day_of_the_races_monday(self):
        # In 2020, Day of the races and hispanity falls on MON
        day_races_2020 = self.cal.get_day_of_the_races(2020)
        self.assertEqual(day_races_2020, date(2020, 10, 12))
        # In 2021, It does not, so it's shifted to the next MON
        day_races_2021 = self.cal.get_day_of_the_races(2021)
        self.assertEqual(day_races_2021, date(2021, 10, 18))

    def test_all_saints_monday(self):
        # In 2021, The All Saints falls on MON
        all_saints_2021 = self.cal.get_all_saints(2021)
        self.assertEqual(all_saints_2021, date(2021, 11, 1))
        # In 2022, It does not, so it's shifted to the next MON
        all_saints_2022 = self.cal.get_all_saints(2022)
        self.assertEqual(all_saints_2022, date(2022, 11, 7))

    def test_cartagena_independence_monday(self):
        # In 2019, The Cartagena Independance falls on MON
        cartagena_2019 = self.cal.get_cartagena_independence(2019)
        self.assertEqual(cartagena_2019, date(2019, 11, 11))
        # In 2020, It does not, so it's shifted to the next MON
        cartagena_2020 = self.cal.get_cartagena_independence(2020)
        self.assertEqual(cartagena_2020, date(2020, 11, 16))


class MexicoTest(GenericCalendarTest):
    cal_class = Mexico

    def test_holidays_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)
        self.assertIn(date(2013, 2, 4), holidays)  # Constitution day
        self.assertIn(date(2013, 3, 18), holidays)  # Benito Juárez's birthday
        self.assertIn(date(2013, 5, 1), holidays)  # Labour day
        self.assertIn(date(2013, 9, 16), holidays)  # Independence day
        self.assertIn(date(2013, 11, 18), holidays)  # Revolution day
        self.assertIn(date(2013, 12, 25), holidays)  # XMas

    def test_shift_to_monday(self):
        observed = set(map(self.cal.get_observed_date, self.cal.holidays_set(2017)))
        # New year on Sunday -> shift
        assert date(2017, 1, 2) in observed
        observed = set(map(self.cal.get_observed_date, self.cal.holidays_set(2016)))
        # XMas on sunday -> shift to monday
        assert date(2016, 12, 26) in observed
        # Same for Labour day
        assert date(2016, 5, 2) in observed

    def test_shift_to_friday(self):
        holidays = self.cal.holidays_set(2021) | self.cal.holidays_set(2022)
        observed = set(map(self.cal.get_observed_date, holidays))
        # January 1st 2022 is a saturday, so we shift to friday
        assert date(2021, 12, 31) in observed
        # Same for Labour day
        assert date(2021, 4, 30) in observed
        holidays = self.cal.holidays_set(2021)
        observed = set(map(self.cal.get_observed_date, holidays))
        # December 25th, 2022 is a saturday, so we shift to friday
        assert date(2021, 12, 24) in observed


class PanamaTest(GenericCalendarTest):
    cal_class = Panama

    def test_holidays_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)
        self.assertIn(date(2013, 1, 9), holidays)  # Martyrs day
        self.assertIn(date(2013, 2, 12), holidays)  # carnival tuesday
        self.assertIn(date(2013, 3, 29), holidays)  # good friday
        self.assertIn(date(2013, 3, 30), holidays)  # easter saturday
        self.assertIn(date(2013, 3, 31), holidays)  # easter sunday
        self.assertIn(date(2013, 5, 1), holidays)  # labour day
        self.assertIn(date(2013, 11, 3), holidays)  # independence day
        self.assertIn(date(2013, 11, 5), holidays)  # colon day
        # Shout in Villa de los Santos
        self.assertIn(date(2013, 11, 10), holidays)
        self.assertIn(date(2013, 11, 28), holidays)  # Independence from spain
        self.assertIn(date(2013, 12, 8), holidays)  # mother day
        self.assertIn(date(2013, 12, 25), holidays)  # XMas


class ParaguayTest(GenericCalendarTest):
    cal_class = Paraguay

    def test_holidays_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 1, 1), holidays)
        self.assertIn(date(2019, 3, 1), holidays)  # Heroes day
        self.assertIn(date(2019, 4, 18), holidays)  # Maundy thursday
        self.assertIn(date(2019, 4, 19), holidays)  # Good friday
        self.assertIn(date(2019, 5, 1), holidays)  # Labour day
        self.assertIn(date(2019, 5, 14), holidays)  # Independance day
        self.assertIn(date(2019, 6, 12), holidays)  # Chaco Armistice Day
        self.assertIn(date(2019, 8, 15), holidays)  # Founding of Asunción
        self.assertIn(date(2019, 9, 29), holidays)  # Boqueron Battle Victory
        self.assertIn(date(2019, 12, 8), holidays)  # Virgin of Caacupe
        self.assertIn(date(2019, 12, 25), holidays)  # XMas

    def test_holidays_2017(self):
        holidays = self.cal.holidays_set(2017)
        # In 2017, Heroes day has been moved to February 27th
        self.assertNotIn(date(2017, 3, 1), holidays)
        self.assertIn(date(2017, 2, 27), holidays)
        # Fundation of Asunción day: moved to August 14 for 2017
        self.assertNotIn(date(2017, 8, 15), holidays)
        self.assertIn(date(2017, 8, 14), holidays)
        # Boqueron Battle Victory Day: moved to October 2nd for 2017
        self.assertNotIn(date(2017, 9, 29), holidays)
        self.assertIn(date(2017, 10, 2), holidays)

    def test_immaculate_conception_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        label = holidays[date(2020, 12, 8)]
        self.assertEqual(label, "Virgin of Caacupé Day")


class BarbadosTest(GenericCalendarTest):
    cal_class = Barbados

    def test_holidays_2009(self):
        holidays = self.cal.holidays_set(2009)
        self.assertIn(date(2009, 1, 1), holidays)
        self.assertIn(date(2009, 1, 21), holidays)  # Errol Barrow Day
        self.assertIn(date(2009, 4, 10), holidays)  # Good Friday
        self.assertIn(date(2009, 4, 12), holidays)  # Easter Sunday
        self.assertIn(date(2009, 4, 13), holidays)  # Easter Monday
        self.assertIn(date(2009, 4, 28), holidays)  # National Heroes Day
        self.assertIn(date(2009, 5, 1), holidays)  # Labour Day
        self.assertIn(date(2009, 6, 1), holidays)  # Whit Monday
        self.assertIn(date(2009, 8, 1), holidays)  # Emancipation Day
        self.assertIn(date(2009, 8, 3), holidays)  # Kabooment Day
        self.assertIn(date(2009, 11, 30), holidays)  # Independant Day
        self.assertIn(date(2009, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2009, 12, 26), holidays)  # Boxing Day

    def test_holidays_2016(self):
        holidays = self.cal.holidays_set(2016)
        self.assertIn(date(2016, 1, 1), holidays)
        self.assertIn(date(2016, 1, 21), holidays)  # Errol Barrow Day
        self.assertIn(date(2016, 3, 25), holidays)  # Good Friday
        self.assertIn(date(2016, 3, 28), holidays)  # Easter Monday
        self.assertIn(date(2016, 4, 28), holidays)  # National Heroes Day
        self.assertIn(date(2016, 5, 2), holidays)  # Labour Day
        self.assertIn(date(2016, 5, 16), holidays)  # Whit Monday
        self.assertIn(date(2016, 8, 1), holidays)  # Kabooment Day
        self.assertIn(date(2016, 8, 2), holidays)  # Emancipation Day
        self.assertIn(date(2016, 11, 30), holidays)  # Independant Day
        self.assertIn(date(2016, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2016, 12, 26), holidays)  # Boxing Day
        self.assertIn(date(2016, 12, 27), holidays)  # Public Holiday

    def test_holidays_2018(self):
        holidays = self.cal.holidays_set(2018)
        self.assertIn(date(2018, 1, 1), holidays)
        self.assertIn(date(2018, 1, 21), holidays)  # Errol Barrow Day
        self.assertIn(date(2018, 1, 22), holidays)  # Errol Barrow Day (shift)
        self.assertIn(date(2018, 3, 30), holidays)  # Good Friday
        self.assertIn(date(2018, 4, 1), holidays)  # Easter Sunday
        self.assertIn(date(2018, 4, 2), holidays)  # Easter Monday
        self.assertIn(date(2018, 4, 28), holidays)  # National Heroes Day
        self.assertIn(date(2018, 5, 1), holidays)  # Labour Day
        self.assertIn(date(2018, 5, 21), holidays)  # Whit Monday
        self.assertIn(date(2018, 8, 1), holidays)  # Emancipation Day
        self.assertIn(date(2018, 8, 6), holidays)  # Kabooment Day
        self.assertIn(date(2018, 11, 30), holidays)  # Independant Day
        self.assertIn(date(2018, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2018, 12, 26), holidays)  # Boxing Day

    def test_holidays_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 1, 1), holidays)
        self.assertIn(date(2019, 1, 21), holidays)  # Errol Barrow Day
        self.assertIn(date(2019, 4, 19), holidays)  # Good Friday
        self.assertIn(date(2019, 4, 21), holidays)  # Easter Sunday
        self.assertIn(date(2019, 4, 22), holidays)  # Easter Monday

        # National Heroes Day & shift
        self.assertIn(date(2019, 4, 28), holidays)
        self.assertIn(date(2019, 4, 29), holidays)  # shft'd

        self.assertIn(date(2019, 5, 1), holidays)  # Labour Day
        self.assertIn(date(2019, 6, 10), holidays)  # Whit Monday
        self.assertIn(date(2019, 8, 1), holidays)  # Emancipation Day
        self.assertIn(date(2019, 8, 5), holidays)  # Kabooment Day
        self.assertIn(date(2019, 11, 30), holidays)  # Independant Day
        self.assertIn(date(2019, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2019, 12, 26), holidays)  # Boxing Day

    def test_holidays_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 1, 1), holidays)
        self.assertIn(date(2020, 1, 21), holidays)  # Errol Barrow Day
        self.assertIn(date(2020, 4, 10), holidays)  # Good Friday
        self.assertIn(date(2020, 4, 13), holidays)  # Easter Monday

        # National Heroes Day & shift
        self.assertIn(date(2020, 4, 28), holidays)

        self.assertIn(date(2020, 5, 1), holidays)  # Labour Day
        self.assertIn(date(2020, 6, 1), holidays)  # Whit Monday
        self.assertIn(date(2020, 8, 1), holidays)  # Emancipation Day
        self.assertIn(date(2020, 8, 3), holidays)  # Kabooment Day
        self.assertIn(date(2020, 11, 30), holidays)  # Independant Day
        self.assertIn(date(2020, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2020, 12, 26), holidays)  # Boxing Day

    def test_holidays_2021(self):
        holidays = self.cal.holidays_set(2021)
        self.assertIn(date(2021, 1, 1), holidays)  # New Year's Day
        self.assertIn(date(2021, 1, 4), holidays)  # Public Holiday
        self.assertIn(date(2021, 1, 5), holidays)  # Public Holiday
        self.assertIn(date(2021, 1, 21), holidays)  # Errol Barrow Day
        self.assertIn(date(2021, 4, 2), holidays)  # Good Friday
        self.assertIn(date(2021, 4, 5), holidays)  # Easter Monday
        self.assertIn(date(2021, 4, 28), holidays)  # National Heroes Day

        self.assertIn(date(2021, 5, 1), holidays)  # Labour Day
        self.assertIn(date(2021, 5, 24), holidays)  # Whit Monday
        self.assertIn(date(2021, 8, 2), holidays)  # Kabooment Day
        self.assertIn(date(2021, 8, 3), holidays)  # Emancipation Day
        self.assertIn(date(2021, 11, 30), holidays)  # Independant Day
        self.assertIn(date(2021, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2021, 12, 27), holidays)  # Boxing Day



from datetime import date

from ..core import ChineseNewYearCalendar, cleaned_date
from ..astronomy import solar_term
from ..registry_tools import iso_register


@iso_register('TW')
class Taiwan(ChineseNewYearCalendar):
    "Taiwan (Republic of China)"
    FIXED_HOLIDAYS = ChineseNewYearCalendar.FIXED_HOLIDAYS + (
        (2, 28, "228 Peace Memorial Day"),
        (4, 4, "Combination of Women's Day and Children's Day"),
        (10, 10, "National Day/Double Tenth Day"),
    )
    include_chinese_new_year_eve = True
    include_chinese_second_day = True

    def is_working_day(self, day, *args, **kwargs):
        day = cleaned_date(day)
        if day in (date(2021, 2, 20), date(2021, 9, 11)):
            return True
        return super().is_working_day(day, *args, **kwargs)

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        # Qingming begins when the sun reaches the celestial
        # longitude of 15° (usually around April 4th or 5th)
        qingming = solar_term(year, 15, 'Asia/Taipei')

        days.extend([
            (
                ChineseNewYearCalendar.lunar(year, 1, 3),
                "Chinese New Year (3rd day)"
            ),
            (qingming, "Qingming Festival"),
            (ChineseNewYearCalendar.lunar(year, 5, 5), "Dragon Boat Festival"),
            (ChineseNewYearCalendar.lunar(year, 8, 15), "Mid-Autumn Festival"),
        ])
        return days



from .barbados import Barbados
from .brazil import (
    Brazil, BrazilAcre, BrazilAlagoas, BrazilAmapa, BrazilAmazonas,
    BrazilBahia, BrazilCeara, BrazilDistritoFederal, BrazilEspiritoSanto,
    BrazilGoias, BrazilMaranhao, BrazilMinasGerais, BrazilMatoGrosso,
    BrazilMatoGrossoDoSul, BrazilPara, BrazilParaiba, BrazilPernambuco,
    BrazilPiaui, BrazilParana, BrazilRioDeJaneiro, BrazilRioGrandeDoNorte,
    BrazilRioGrandeDoSul, BrazilRondonia, BrazilRoraima, BrazilSantaCatarina,
    BrazilSaoPauloState, BrazilSaoPauloCity, BrazilSergipe, BrazilTocantins,
    BrazilVitoriaCity, BrazilVilaVelhaCity, BrazilCariacicaCity,
    BrazilGuarapariCity, BrazilSerraCity, BrazilRioBrancoCity,
    BrazilMaceioCity, BrazilManausCity, BrazilMacapaCity, BrazilSalvadorCity,
    BrazilFortalezaCity, BrazilGoianiaCity, BrazilBeloHorizonteCity,
    BrazilCampoGrandeCity, BrazilCuiabaCity, BrazilBelemCity,
    BrazilJoaoPessoaCity, BrazilRecifeCity, BrazilTeresinaCity,
    BrazilCuritibaCity, BrazilNatalCity, BrazilPortoVelhoCity,
    BrazilBoaVistaCity, BrazilPortoAlegreCity, BrazilChapecoCity,
    BrazilFlorianopolisCity, BrazilJoinvilleCity, BrazilAracajuCity,
    BrazilSorocabaCity, BrazilPalmasCity, BrazilBankCalendar
)
from .canada import (
    Canada, Ontario, Quebec, BritishColumbia, Alberta, Saskatchewan, Manitoba,
    NewBrunswick, NovaScotia, PrinceEdwardIsland, Newfoundland, Yukon,
    NorthwestTerritories, Nunavut
)
from .chile import Chile
from .colombia import Colombia
from .mexico import Mexico
from .panama import Panama
from .paraguay import Paraguay
from .argentina import Argentina


__all__ = (
    # Brazil & its states.
    'Brazil',
    'BrazilAcre',
    'BrazilAlagoas',
    'BrazilAmapa',
    'BrazilAmazonas',
    'BrazilBahia',
    'BrazilCeara',
    'BrazilDistritoFederal',
    'BrazilEspiritoSanto',
    'BrazilGoias',
    'BrazilMaranhao',
    'BrazilMinasGerais',
    'BrazilMatoGrosso',
    'BrazilMatoGrossoDoSul',
    'BrazilPara',
    'BrazilParaiba',
    'BrazilPernambuco',
    'BrazilPiaui',
    'BrazilParana',
    'BrazilRioDeJaneiro',
    'BrazilRioGrandeDoNorte',
    'BrazilRioGrandeDoSul',
    'BrazilRondonia',
    'BrazilRoraima',
    'BrazilSantaCatarina',
    'BrazilSaoPauloState',
    'BrazilSaoPauloCity',
    'BrazilSergipe',
    'BrazilTocantins',
    'BrazilVitoriaCity',
    'BrazilVilaVelhaCity',
    'BrazilCariacicaCity',
    'BrazilGuarapariCity',
    'BrazilSerraCity',
    'BrazilBankCalendar',
    'BrazilRioBrancoCity',
    'BrazilMaceioCity',
    'BrazilManausCity',
    'BrazilMacapaCity',
    'BrazilSalvadorCity',
    'BrazilFortalezaCity',
    'BrazilGoianiaCity',
    'BrazilBeloHorizonteCity',
    'BrazilCampoGrandeCity',
    'BrazilCuiabaCity',
    'BrazilBelemCity',
    'BrazilJoaoPessoaCity',
    'BrazilRecifeCity',
    'BrazilTeresinaCity',
    'BrazilCuritibaCity',
    'BrazilNatalCity',
    'BrazilPortoVelhoCity',
    'BrazilBoaVistaCity',
    'BrazilPortoAlegreCity',
    'BrazilChapecoCity',
    'BrazilFlorianopolisCity',
    'BrazilJoinvilleCity',
    'BrazilAracajuCity',
    'BrazilSorocabaCity',
    'BrazilPalmasCity',
    # Canada
    'Canada',
    'Ontario',
    'Quebec',
    'BritishColumbia',
    'Alberta',
    'Saskatchewan',
    'Manitoba',
    'NewBrunswick',
    'NovaScotia',
    'PrinceEdwardIsland',
    'Newfoundland',
    'Yukon',
    'NorthwestTerritories',
    'Nunavut',
    # Other american countries
    'Barbados',
    'Chile',
    'Colombia',
    'Mexico',
    'Panama',
    'Paraguay',
    'Argentina',
)



from copy import copy
from datetime import timedelta, date

from ..core import OrthodoxCalendar, SAT, SUN
from ..registry_tools import iso_register


@iso_register('BG')
class Bulgaria(OrthodoxCalendar):
    'Bulgaria'

    FIXED_HOLIDAYS = OrthodoxCalendar.FIXED_HOLIDAYS + (
        (3, 3, "Liberation Day"),  # Ден на Освобождението на Б
        (5, 6, "Saint George's Day"),  # Гергьовден, ден на храброс
        (5, 24, "Saints Cyril & Methodius Day"),  # Ден на българската просвет
        (9, 6, "Unification Day"),  # Ден на Съединението
        (9, 22, "Independence Day"),  # Ден на независимостта на Б
    )

    # Civil holidays
    include_labour_day = True
    # Ден на труда и на междунар
    labour_day_label = "International Workers' Day"

    # Christian holidays
    include_good_friday = True
    include_easter_saturday = True
    include_easter_sunday = True
    include_easter_monday = True
    include_christmas_eve = True  # Бъдни вечер
    include_christmas = True  # Рождество Христово
    include_boxing_day = True
    # despite being an Orthodox calendar, don't observe Orthodox XMas
    include_orthodox_christmas = False

    # wikipedia says The Bulgarians have two days of Christmas,
    # both called Christmas Day
    boxing_day_label = "Christmas"

    def get_shifted_holidays(self, days):
        for holiday, label in days:
            if holiday.weekday() == SUN:
                yield (
                    holiday + timedelta(days=1),
                    f'{label} shift'
                )
            elif holiday.weekday() == SAT:
                yield (
                    holiday + timedelta(days=2),
                    f'{label} shift'
                )

    def get_fixed_holidays(self, year):
        """
        Return fixed holidays, with shifts computed accordingly.
        """
        # 2021 exception.
        # Because May 1st is both International Workers' day and Easter
        self.include_labour_day = (year != 2021)

        # Unshifted days are here:
        days = super().get_fixed_holidays(year)
        days_to_inspect = copy(days)
        for day_shifted in self.get_shifted_holidays(days_to_inspect):
            days.append(day_shifted)

        # 2021 exception.
        # Because May 1st is both International Workers' day and Easter
        if year == 2021:
            days.append((date(2021, 5, 4), self.labour_day_label))
        return days

    def shift_christmas_boxing_days(self, year):
        """
        Return Christmas shifts.

        They usually follow the same rule as used in the UK, but the fact that
        there's a second Christmas holiday requires an override.
        """
        days = super().shift_christmas_boxing_days(year=year)
        # In this case we'll need an extra day for the Second day of XMas
        if date(year, 12, 25).weekday() == SUN:
            days.append(
                (date(year, 12, 28), "Christmas (in lieu)")
            )
        return days

    def get_variable_days(self, year):
        """
        Return variable holidays, with Christmas shifts computed accordingly.
        """
        days = super().get_variable_days(year)
        # Boxing day & XMas shift
        shifts = self.shift_christmas_boxing_days(year=year)
        days.extend(shifts)
        return days



from ..core import WesternCalendar
from ..registry_tools import iso_register


@iso_register('MZ')
class Mozambique(WesternCalendar):
    "Mozambique"
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (2, 3, "Mozambican Heroes' Day"),
        (4, 7, "Mozambican Women's Day"),
        (6, 25, "Independence Day"),
        (9, 7, "Victory Day"),
        (9, 25, "Armed Forces Day"),
        (10, 4, "Peace And Reconciliation Day"),
    )

    # Civil holidays
    include_labour_day = True

    # Christian holidays
    include_good_friday = True
    include_christmas = True



from datetime import timedelta, date

from ..core import WesternCalendar, MON
from ..registry_tools import iso_register


@iso_register('CO')
class Colombia(WesternCalendar):
    "Colombia"
    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (7, 20, "Independence Day"),
        (8, 7, "Boyacá Battle"),
    )
    # Civil holidays
    include_labour_day = True
    # Christian holidays
    include_palm_sunday = True
    include_holy_thursday = True
    include_good_friday = True
    include_easter_sunday = True
    include_corpus_christi = True
    include_immaculate_conception = True

    def get_epiphany(self, year):
        """
        Epiphany is shifted in Colombia
        """
        base_day = date(year, 1, 6)
        return self.get_first_weekday_after(base_day, MON)

    def get_saint_joseph(self, year):
        base_day = date(year, 3, 19)
        return self.get_first_weekday_after(base_day, MON)

    def get_ascension(self, year):
        # By default, Ascension falls on THU.
        # But in Colombia, they celebrate it the next MON.
        base_day = self.get_ascension_thursday(year)
        return self.get_first_weekday_after(base_day, MON)

    def get_corpus_christi(self, year):
        # By default, Corpus Christi falls 60 days after Easter.
        # But in Colombia, they celebrate it the next MON.
        base_day = super().get_corpus_christi(year)
        return self.get_first_weekday_after(base_day, MON)

    def get_sacred_heart(self, year):
        # By default, Sacred Heart falls 68 days after Easter.
        # But in Colombia, they celebrate it the next MON.
        base_day = self.get_easter_sunday(year) + timedelta(days=68)
        return self.get_first_weekday_after(base_day, MON)

    def get_saint_peter_and_saint_paul(self, year):
        base_day = date(year, 6, 29)
        return self.get_first_weekday_after(base_day, MON)

    def get_assumption(self, year):
        # By default, Assumption is a fixed date (August 15th)
        # But in Colombia, they celebrate it the next MON.
        base_day = date(year, 8, 15)
        return self.get_first_weekday_after(base_day, MON)

    def get_day_of_the_races(self, year):
        """
        Return Day of the Races and Hispanity

        a.k.a. "Día de la Raza"
        Fixed to the next MON after October 12th (Columbus Day)
        """
        base_day = date(year, 10, 12)
        return self.get_first_weekday_after(base_day, MON)

    def get_all_saints(self, year):
        # By default, All Saints is a fixed date (November 1st)
        # But in Colombia, they celebrate it the next MON.
        base_day = date(year, 11, 1)
        return self.get_first_weekday_after(base_day, MON)

    def get_cartagena_independence(self, year):
        """
        Cartagena independance day

        Fixed to the next MON after November 11th.
        """
        base_day = date(year, 11, 11)
        return self.get_first_weekday_after(base_day, MON)

    def get_variable_days(self, year):
        """
        Return variable holidays for Colombia

        The following days are set to "the next MON after the 'true' date".

        * Epiphany,
        * Saint Joseph,
        * Ascension,
        * Corpus Christi,
        * Sacred Heart,
        * Saint Peter & Saint Paul
        * Assumption
        * Columbus Day / Race Day
        * All Saints
        * Cartagena Independance
        """
        days = super().get_variable_days(year)
        days.extend([
            (self.get_epiphany(year), "Epiphany"),
            (self.get_saint_joseph(year), "Saint Joseph"),
            (self.get_ascension(year), "Ascension"),
            (self.get_sacred_heart(year), "Sacred Heart"),
            (self.get_saint_peter_and_saint_paul(year),
                "Saint Peter and Saint Paul"),
            (self.get_assumption(year), "Assumption of Mary to Heaven"),
            (self.get_day_of_the_races(year),
                "Day of the Races and Hispanity"),
            (self.get_all_saints(year), "All Saints"),
            (self.get_cartagena_independence(year),
                "Cartagena's Independence"),
        ])

        return days



from ..registry_tools import iso_register
from .core import UnitedStates


@iso_register('US-WY')
class Wyoming(UnitedStates):
    """Wyoming"""
    martin_luther_king_label = "Martin Luther King, Jr. / Wyoming Equality Day"



"""
Victoria Day mixins
"""
from ....core import MON


class VictoriaDayMixin:
    include_victoria_day = True
    victoria_day_label = "Victoria Day"


class VictoriaDayFourthMondayMay(VictoriaDayMixin):

    def get_victoria_day(self, year):
        return (
            self.get_nth_weekday_in_month(year, 5, MON, 4),
            self.victoria_day_label
        )


class VictoriaDayLastMondayMay(VictoriaDayMixin):

    def get_victoria_day(self, year):
        return (
            self.get_last_weekday_in_month(year, 5, MON),
            self.victoria_day_label
        )


class VictoriaDayFirstMondayJune(VictoriaDayMixin):

    def get_victoria_day(self, year):
        return (
            self.get_nth_weekday_in_month(year, 6, MON),
            self.victoria_day_label
        )



from .core import UnitedStates
from ..registry_tools import iso_register


@iso_register('US-AR')
class Arkansas(UnitedStates):
    """Arkansas"""
    include_christmas_eve = True
    presidents_day_label = ("George Washington's Birthday"
                            " and Daisy Gatson Bates Day")
    include_columbus_day = False



import tempfile
from pathlib import Path

from ..core import Calendar
from ..exceptions import ICalExportRangeError, ICalExportTargetPathError

from . import CoreCalendarTest


class FakeCalendar(Calendar):
    """Fake Calendar"""


class IcalExportPeriodTest(CoreCalendarTest):
    cal_class = FakeCalendar

    def test_empty_range(self):
        # No period means 2000-2030
        result = self.cal._get_ical_period()
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2000)
        self.assertEqual(result[1], 2030)
        # No period means 2000-2030
        result = self.cal._get_ical_period(period=())
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2000)
        self.assertEqual(result[1], 2030)
        # No period means 2000-2030
        result = self.cal._get_ical_period(period=[])
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2000)
        self.assertEqual(result[1], 2030)

    def test_range_lists(self):
        # Period: 2010-2020
        result = self.cal._get_ical_period(period=[2010, 2020])
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2010)
        self.assertEqual(result[1], 2020)

        # Bad period: 2020-2010 => Swap terms
        result = self.cal._get_ical_period(period=[2020, 2010])
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2010)
        self.assertEqual(result[1], 2020)

        # Too long of a list => take the extremes
        result = self.cal._get_ical_period(period=[2010, 2015, 2020])
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2010)
        self.assertEqual(result[1], 2020)

        # Assume that the range is only one year
        result = self.cal._get_ical_period(period=[2020, 2020])
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2020)
        self.assertEqual(result[1], 2020)

    def test_range_tuples(self):
        # Period: 2010-2020
        result = self.cal._get_ical_period(period=(2010, 2020))
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2010)
        self.assertEqual(result[1], 2020)

        # Bad period: 2020-2010 => Swap terms
        result = self.cal._get_ical_period(period=(2020, 2010))
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2010)
        self.assertEqual(result[1], 2020)

        # Too long of a list => take the extremes
        result = self.cal._get_ical_period(period=(2010, 2015, 2020))
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2010)
        self.assertEqual(result[1], 2020)

        # Assume that the range is only one year
        result = self.cal._get_ical_period(period=(2020, 2020))
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 2020)
        self.assertEqual(result[1], 2020)

    def test_bad_ranges(self):
        # Bad period: String
        with self.assertRaises(ICalExportRangeError):
            self.cal._get_ical_period(period="something")

        # Bad period: Dict
        with self.assertRaises(ICalExportRangeError):
            self.cal._get_ical_period(period={"a": 2000, "b": "foobar"})

        # Bad period: List of non-integers
        with self.assertRaises(ICalExportRangeError):
            self.cal._get_ical_period(period=["foo", "bar"])

        # Bad period: Tuple of non-integers
        with self.assertRaises(ICalExportRangeError):
            self.cal._get_ical_period(period=("foo", "bar"))


class ICalExportTargetPath(CoreCalendarTest):
    cal_class = FakeCalendar

    def test_empty_path(self):
        with self.assertRaises(ICalExportTargetPathError):
            self.cal._get_ical_target_path("")

        with self.assertRaises(ICalExportTargetPathError):
            self.cal._get_ical_target_path(None)

    def test_target_is_directory(self):
        temp_dir = Path(tempfile.gettempdir()) / "failed_ical_tests"
        temp_dir.mkdir(parents=True, exist_ok=True)

        with self.assertRaises(ICalExportTargetPathError):
            self.cal._get_ical_target_path(temp_dir)

    def test_no_extension(self):
        # Relative paths
        self.assertEqual(self.cal._get_ical_target_path("a"), Path("a.ics"))
        self.assertEqual(
            self.cal._get_ical_target_path("a/b"),
            Path("a/b.ics")
        )
        # Absolute path
        self.assertEqual(
            self.cal._get_ical_target_path("/path/to/a"),
            Path("/path/to/a.ics")
        )

    def test_known_extensions(self):
        for ext in ('ical', 'ics', 'ifb', 'icalendar'):
            filename = Path(f"a.{ext}")
            self.assertEqual(
                self.cal._get_ical_target_path(filename),
                filename
            )

    def test_added_extensions(self):
        self.assertEqual(self.cal._get_ical_target_path('a.'), Path('a..ics'))
        self.assertEqual(
            self.cal._get_ical_target_path('a.txt'),
            Path('a.txt.ics')
        )
        self.assertEqual(
            self.cal._get_ical_target_path('.test'),
            Path('.test.ics')
        )



import pytest
from datetime import date
from math import pi
from unittest.mock import patch

from skyfield.api import Loader
from skyfield_data import get_skyfield_data_path

from ..astronomy import calculate_equinoxes, solar_term, newton_angle_function


def test_calculate_some_equinoxes():
    assert calculate_equinoxes(2010) == (date(2010, 3, 20), date(2010, 9, 23))
    assert calculate_equinoxes(2010, 'Asia/Taipei') == (
        date(2010, 3, 21), date(2010, 9, 23)
    )
    assert calculate_equinoxes(2013) == (date(2013, 3, 20), date(2013, 9, 22))
    assert calculate_equinoxes(2014) == (date(2014, 3, 20), date(2014, 9, 23))
    assert calculate_equinoxes(2020) == (date(2020, 3, 20), date(2020, 9, 22))


def test_qingming_festivals():
    assert solar_term(2001, 15) == date(2001, 4, 4)
    assert solar_term(2001, 15, 'Asia/Taipei') == date(2001, 4, 5)
    assert solar_term(2011, 15) == date(2011, 4, 5)
    assert solar_term(2014, 15) == date(2014, 4, 4)
    assert solar_term(2016, 15) == date(2016, 4, 4)
    assert solar_term(2017, 15) == date(2017, 4, 4)


def test_qingming_festivals_hk():
    assert solar_term(2018, 15, 'Asia/Hong_Kong') == date(2018, 4, 5)
    assert solar_term(2019, 15, 'Asia/Hong_Kong') == date(2019, 4, 5)
    assert solar_term(2020, 15, 'Asia/Hong_Kong') == date(2020, 4, 4)
    assert solar_term(2021, 15, 'Asia/Hong_Kong') == date(2021, 4, 4)


@pytest.fixture(scope='session')
def params_newton_angle():
    """
    Session-scoped fixture to "cache" the newton angle func parameters
    """
    load = Loader(get_skyfield_data_path())
    ts = load.timescale()
    planets = load('de421.bsp')
    earth = planets['earth']
    sun = planets['sun']

    jan_first = ts.utc(date(2021, 1, 1))
    t0 = ts.tt_jd(jan_first.tt).tt
    return t0, ts, earth, sun


def test_newton_angle_function_normal_range(params_newton_angle):
    """
    Test the newton angle func when the longitude is in the range [-pi, +pi].
    """
    t0, ts, earth, sun = params_newton_angle

    with patch('calendra.astronomy.get_current_longitude') as patched:
        patched.return_value = pi
        assert newton_angle_function(t0, ts, 0, earth, sun) == -pi


def test_newton_angle_function_above_pi(params_newton_angle):
    """
    Test the newton angle function when the longitude is > +pi.

    This should not happen, but it was implemented in the function, so in order
    to make sure that the resulting angle is always in the range [-pi, +pi],
    we added these tests with those out of range values.
    """
    t0, ts, earth, sun = params_newton_angle

    with patch('calendra.astronomy.get_current_longitude') as patched:
        patched.return_value = pi + 0.1
        expected = pytest.approx(-.1)
        assert newton_angle_function(t0, ts, 0, earth, sun) == expected


def test_newton_angle_function_below_minus_pi(params_newton_angle):
    """
    Test the newton angle function when the longitude is < -pi.

    This should not happen, but it was implemented in the function, so in order
    to make sure that the resulting angle is always in the range [-pi, +pi],
    we added these tests with those out of range values.
    """
    t0, ts, earth, sun = params_newton_angle

    with patch('calendra.astronomy.get_current_longitude') as patched:
        patched.return_value = -pi - 0.1
        expected = pytest.approx(.1)
        assert newton_angle_function(t0, ts, 0, earth, sun) == expected



from unittest.mock import patch
from datetime import date
import time
from . import GenericCalendarTest

import pytest

from ..asia import (
    HongKong, HongKongBank,
    Japan, JapanBank, Qatar, Singapore,
    SouthKorea, Taiwan, Malaysia, China, Israel
)
from ..asia.china import holidays as china_holidays
from ..exceptions import CalendarError


class ChinaTest(GenericCalendarTest):

    cal_class = China

    def test_year_2018(self):
        holidays = self.cal.holidays_set(2018)
        self.assertIn(date(2018, 1, 1), holidays)    # New year
        self.assertIn(date(2018, 2, 15), holidays)   # Spring Festival
        self.assertIn(date(2018, 2, 21), holidays)   # Spring Festival
        self.assertIn(date(2018, 10, 1), holidays)   # National Day
        self.assertIn(date(2018, 10, 7), holidays)   # National Day

        # Variable days
        self.assertIn(date(2018, 4, 5), holidays)    # Ching Ming Festival
        self.assertIn(date(2018, 4, 7), holidays)    # Ching Ming Festival
        self.assertIn(date(2018, 4, 29), holidays)   # Labour Day Holiday
        self.assertIn(date(2018, 5, 1), holidays)    # Labour Day Holiday
        self.assertIn(date(2018, 6, 18), holidays)   # Dragon Boat Festival
        self.assertIn(date(2018, 9, 24), holidays)   # Mid-Autumn Festival
        self.assertIn(date(2018, 12, 30), holidays)  # New year
        self.assertIn(date(2018, 12, 31), holidays)  # New year

    def test_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertNotIn(date(2019, 4, 28), holidays)  # Labour Day Shift
        self.assertIn(date(2019, 5, 1), holidays)  # Labour Day Holiday
        self.assertIn(date(2019, 5, 2), holidays)  # Labour Day Holiday
        self.assertIn(date(2019, 5, 3), holidays)  # Labour Day Holiday
        self.assertNotIn(date(2019, 5, 5), holidays)  # Labour Day Shift

    def test_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 1, 1), holidays)  # New Year
        self.assertIn(date(2020, 1, 25), holidays)  # Spring Festival
        self.assertIn(date(2020, 4, 4), holidays)  # Ching Ming Festival
        self.assertIn(date(2020, 4, 6), holidays)  # Ching Ming Festival
        self.assertIn(date(2020, 5, 1), holidays)  # Labour Day Holiday
        self.assertIn(date(2020, 5, 5), holidays)  # Labour Day Holiday
        self.assertIn(date(2020, 6, 25), holidays)  # Dragon Boat Festival
        self.assertIn(date(2020, 6, 27), holidays)  # Dragon Boat Festival
        self.assertIn(date(2020, 10, 1), holidays)  # National Day
        self.assertIn(date(2020, 10, 8), holidays)  # National Day

        self.assertNotIn(date(2020, 1, 19), holidays)  # Spring Festival Shift
        self.assertNotIn(date(2020, 2, 1), holidays)  # Spring Festival Shift
        self.assertNotIn(date(2020, 4, 26), holidays)  # Labour Day Shift
        self.assertNotIn(date(2020, 5, 9), holidays)  # Labour Day Shift
        self.assertNotIn(date(2020, 6, 28), holidays)  # Dragon Boat Shift
        self.assertNotIn(date(2020, 9, 27), holidays)  # National Day Shift
        self.assertNotIn(date(2020, 10, 10), holidays)  # National Day Shift

    def test_year_2021(self):
        holidays = self.cal.holidays_set(2021)
        self.assertIn(date(2021, 1, 1), holidays)  # New Year
        self.assertIn(date(2021, 2, 11), holidays)  # Spring Festival
        self.assertIn(date(2021, 2, 17), holidays)  # Spring Festival
        self.assertIn(date(2021, 4, 3), holidays)  # Ching Ming Festival
        self.assertIn(date(2021, 4, 5), holidays)  # Ching Ming Festival
        self.assertIn(date(2021, 5, 1), holidays)  # Labour Day Holiday
        self.assertIn(date(2021, 5, 5), holidays)  # Labour Day Holiday
        self.assertIn(date(2021, 6, 12), holidays)  # Dragon Boat Festival
        self.assertIn(date(2021, 6, 14), holidays)  # Dragon Boat Festival
        self.assertIn(date(2021, 9, 19), holidays)  # Mid-Autumn Festival
        self.assertIn(date(2021, 9, 21), holidays)  # Mid-Autumn Festival
        self.assertIn(date(2021, 10, 1), holidays)  # National Day
        self.assertIn(date(2021, 10, 7), holidays)  # National Day

        self.assertNotIn(date(2021, 2, 7), holidays)  # Spring Festival Shift
        self.assertNotIn(date(2021, 2, 20), holidays)  # Spring Festival Shift
        self.assertNotIn(date(2021, 4, 25), holidays)  # Labour Day Shift
        self.assertNotIn(date(2021, 5, 8), holidays)  # Labour Day Shift
        self.assertNotIn(date(2021, 9, 18), holidays)  # Mid-Autumn Shift
        self.assertNotIn(date(2021, 9, 26), holidays)  # National Day Shift
        self.assertNotIn(date(2021, 10, 9), holidays)  # National Day Shift

    def test_missing_holiday_year(self):
        save_2018 = china_holidays[2018]
        del china_holidays[2018]
        with self.assertRaises(CalendarError):
            self.cal.holidays_set(2018)
        china_holidays[2018] = save_2018

    @pytest.mark.xfail(reason="Fails in 2022", strict=True)
    def test_warning(self):
        year = date.today().year
        with patch('warnings.warn') as patched:
            self.cal.get_calendar_holidays(year)
        patched.assert_called_with(
            'Support years 2018-2021 currently, need update every year.'
        )

    def test_is_working_day(self):
        # It's a SAT, but it's a working day this year
        self.assertTrue(self.cal.is_working_day(date(2019, 2, 2)))
        # It's a SUN, but it's a working day this year
        self.assertTrue(self.cal.is_working_day(date(2019, 2, 3)))

    def test_add_working_days(self):
        # Normally, February 1st + 1 working day would be on February 4th
        # Because 2nd and 3rd are on SAT and SUN.
        self.assertEqual(
            self.cal.add_working_days(date(2019, 2, 1), 1),
            date(2019, 2, 2)
        )

    def test_sub_working_days(self):
        # Normally, February 4th - 1 working day would be on February 1st
        # Because 2nd and 3rd are on SAT and SUN.
        self.assertEqual(
            self.cal.sub_working_days(date(2019, 2, 4), 1),
            date(2019, 2, 3)
        )


class HongKongTest(GenericCalendarTest):

    cal_class = HongKong

    def test_year_2010(self):
        # Interesting because Christmas fell on a Saturday and CNY fell
        # on a Sunday, so didn't roll, and Ching Ming was on the same day
        # as Easter Monday
        # https://www.gov.hk/en/about/abouthk/holiday/2010.htm
        holidays = self.cal.holidays_set(2010)
        self.assertIn(date(2010, 1, 1), holidays)    # New Year
        self.assertIn(date(2010, 2, 13), holidays)   # Chinese new year (shift)
        self.assertIn(date(2010, 2, 15), holidays)   # Chinese new year
        self.assertIn(date(2010, 2, 16), holidays)   # Chinese new year
        self.assertNotIn(date(2010, 2, 17), holidays)  # Not Chinese new year
        self.assertIn(date(2010, 4, 2), holidays)    # Good Friday
        self.assertIn(date(2010, 4, 3), holidays)    # Day after Good Friday
        self.assertIn(date(2010, 4, 5), holidays)    # Easter Monday
        self.assertIn(date(2010, 4, 6), holidays)    # Ching Ming (shifted)
        self.assertIn(date(2010, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2010, 5, 21), holidays)   # Buddha's Birthday
        self.assertIn(date(2010, 6, 16), holidays)   # Tuen Ng Festival
        self.assertIn(date(2010, 7, 1), holidays)    # HK SAR Establishment Day
        self.assertIn(date(2010, 9, 23), holidays)   # Day after Mid-Autumn
        self.assertIn(date(2010, 10, 1), holidays)   # National Day
        self.assertIn(date(2010, 10, 16), holidays)  # Chung Yeung Festival
        self.assertIn(date(2010, 12, 25), holidays)  # Christmas Day
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2010, 12, 27), observed)  # Boxing Day (shifted)

    def test_year_2013(self):
        # https://www.gov.hk/en/about/abouthk/holiday/2013.htm
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)    # New Year
        self.assertIn(date(2013, 2, 11), holidays)   # Chinese new year
        self.assertIn(date(2013, 2, 12), holidays)   # Chinese new year
        self.assertIn(date(2013, 2, 13), holidays)   # Chinese new year
        self.assertIn(date(2013, 3, 29), holidays)   # Good Friday
        self.assertIn(date(2013, 3, 30), holidays)   # Day after Good Friday
        self.assertIn(date(2013, 4, 1), holidays)    # Easter Monday
        self.assertIn(date(2013, 4, 4), holidays)    # Ching Ming
        self.assertIn(date(2013, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2013, 5, 17), holidays)   # Buddha's Birthday
        self.assertIn(date(2013, 6, 12), holidays)   # Tuen Ng Festival
        self.assertIn(date(2013, 7, 1), holidays)    # HK SAR Establishment Day
        self.assertIn(date(2013, 9, 20), holidays)   # Day after Mid-Autumn
        self.assertIn(date(2013, 10, 1), holidays)   # National Day
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2013, 10, 14), observed)  # Chung Yeung Festival
        self.assertIn(date(2013, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2013, 12, 26), holidays)  # Boxing Day

    def test_year_2016(self):
        # https://www.gov.hk/en/about/abouthk/holiday/2016.htm
        holidays = self.cal.holidays_set(2016)
        self.assertIn(date(2016, 1, 1), holidays)    # New Year
        self.assertIn(date(2016, 2, 8), holidays)    # Chinese new year
        self.assertIn(date(2016, 2, 9), holidays)    # Chinese new year
        self.assertIn(date(2016, 2, 10), holidays)   # Chinese new year
        self.assertIn(date(2016, 3, 25), holidays)   # Good Friday
        self.assertIn(date(2016, 3, 26), holidays)   # Day after Good Friday
        self.assertIn(date(2016, 3, 28), holidays)   # Easter Monday
        self.assertIn(date(2016, 4, 4), holidays)    # Ching Ming
        self.assertIn(date(2016, 5, 1), holidays)    # Labour Day (SUN)
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2016, 5, 2), observed)    # Labour Day (shifted)
        self.assertIn(date(2016, 5, 14), holidays)   # Buddha's Birthday
        self.assertIn(date(2016, 6, 9), holidays)    # Tuen Ng Festival
        self.assertIn(date(2016, 7, 1), holidays)    # HK SAR Establishment Day
        self.assertIn(date(2016, 9, 16), holidays)   # Day after Mid-Autumn
        self.assertIn(date(2016, 10, 1), holidays)   # National Day
        self.assertIn(date(2016, 10, 10), observed)  # Chung Yeung Festival
        self.assertIn(date(2016, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2016, 12, 26), holidays)  # Christmas + Boxing Day
        self.assertIn(date(2016, 12, 27), holidays)  # Second weekday

    def test_year_2017(self):
        # https://www.gov.hk/en/about/abouthk/holiday/2017.htm
        holidays = self.cal.holidays_set(2017)
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2017, 1, 2), observed)    # New Year (shifted)
        self.assertIn(date(2017, 1, 28), holidays)   # Chinese new year
        self.assertIn(date(2017, 1, 30), holidays)   # Chinese new year
        self.assertIn(date(2017, 1, 31), holidays)   # Chinese new year
        self.assertIn(date(2017, 4, 4), holidays)    # Ching Ming
        self.assertIn(date(2017, 4, 14), holidays)   # Good Friday
        self.assertIn(date(2017, 4, 15), holidays)   # Day after Good Friday
        self.assertIn(date(2017, 4, 17), holidays)   # Easter Monday
        self.assertIn(date(2017, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2017, 5, 3), holidays)    # Buddha's Birthday
        self.assertIn(date(2017, 5, 30), holidays)   # Tuen Ng Festival
        self.assertIn(date(2017, 7, 1), holidays)    # HK SAR Establishment Day
        self.assertIn(date(2017, 10, 2), observed)   # National Day (shifted)
        self.assertIn(date(2017, 10, 5), holidays)   # Day after Mid-Autumn
        self.assertIn(date(2017, 10, 28), holidays)  # Chung Yeung Festival
        self.assertIn(date(2017, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2017, 12, 26), holidays)  # Boxing Day

    def test_chingming_festival(self):
        # This is the same as the Taiwan test, just different spelling
        # Could move this into a Core test
        self.assertIn(date(2005, 4, 5), self.cal.holidays_set(2005))
        self.assertIn(date(2006, 4, 5), self.cal.holidays_set(2006))
        self.assertIn(date(2007, 4, 5), self.cal.holidays_set(2007))
        self.assertIn(date(2008, 4, 4), self.cal.holidays_set(2008))
        self.assertIn(date(2010, 4, 5), self.cal.holidays_set(2010))
        self.assertIn(date(2011, 4, 5), self.cal.holidays_set(2011))
        self.assertIn(date(2012, 4, 4), self.cal.holidays_set(2012))
        self.assertIn(date(2013, 4, 4), self.cal.holidays_set(2013))
        self.assertIn(date(2014, 4, 5), self.cal.holidays_set(2014))
        self.assertIn(date(2015, 4, 4), self.cal.holidays_set(2015))
        self.assertIn(date(2016, 4, 4), self.cal.holidays_set(2016))
        self.assertIn(date(2017, 4, 4), self.cal.holidays_set(2017))
        self.assertIn(date(2018, 4, 5), self.cal.holidays_set(2018))

    def test_holidays_2020(self):
        # https://www.gov.hk/en/about/abouthk/holiday/2020.htm
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 1, 1), holidays)    # New Year
        self.assertIn(date(2020, 1, 25), holidays)   # Chinese new year
        self.assertIn(date(2020, 1, 27), holidays)   # Chinese new year
        self.assertIn(date(2020, 1, 28), holidays)   # Chinese new year
        self.assertIn(date(2020, 4, 4), holidays)    # Ching Ming
        self.assertIn(date(2020, 4, 10), holidays)   # Good Friday
        self.assertIn(date(2020, 4, 11), holidays)   # Day after Good Friday
        self.assertIn(date(2020, 4, 13), holidays)   # Easter Monday
        self.assertIn(date(2020, 4, 30), holidays)    # Buddha's Birthday
        self.assertIn(date(2020, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2020, 6, 25), holidays)   # Tuen Ng Festival
        self.assertIn(date(2020, 7, 1), holidays)    # HK SAR Establishment Day
        self.assertIn(date(2020, 10, 1), holidays)   # National Day
        self.assertIn(date(2020, 10, 2), holidays)   # Day after Mid-Autumn
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2020, 10, 26), observed)  # Chung Yeung Festival
        self.assertIn(date(2020, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2020, 12, 26), holidays)  # Boxing Day
        # Special: Boxing day is not shifted, because it's not a SUN
        self.assertNotIn(date(2020, 12, 28), holidays)

    def test_holidays_2021(self):
        # https://www.gov.hk/en/about/abouthk/holiday/2021.htm
        holidays = self.cal.holidays_set(2021)
        self.assertIn(date(2021, 1, 1), holidays)    # New Year
        self.assertIn(date(2021, 2, 12), holidays)   # Chinese new year
        self.assertIn(date(2021, 2, 13), holidays)   # Chinese new year
        self.assertIn(date(2021, 2, 14), holidays)   # Chinese new year (SUN)
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2021, 2, 15), observed)   # Chinese new year
        self.assertIn(date(2021, 4, 2), holidays)    # Good Friday
        self.assertIn(date(2021, 4, 3), holidays)    # Day after Good Friday
        self.assertIn(date(2021, 4, 4), holidays)    # Ching Ming
        self.assertIn(date(2021, 4, 5), holidays)    # Ching Ming shift+Easter
        self.assertIn(date(2021, 4, 6), holidays)    # Day after Easter Monday
        self.assertIn(date(2021, 5, 1), holidays)    # Labour Day
        self.assertIn(date(2021, 5, 19), holidays)   # Buddha's Birthday
        self.assertIn(date(2021, 6, 14), holidays)   # Tuen Ng Festival
        self.assertIn(date(2021, 7, 1), holidays)    # HK SAR Establishment Day
        self.assertIn(date(2021, 9, 22), holidays)   # Day after Mid-Autumn
        self.assertIn(date(2021, 10, 1), holidays)   # National Day
        self.assertIn(date(2021, 10, 14), observed)  # Chung Yeung Festival
        self.assertIn(date(2021, 12, 25), holidays)  # Christmas Day
        self.assertIn(date(2021, 12, 27), observed)  # First weekday after Xmas

    def test_are_saturdays_working_days(self):
        # Let's start with february 6th.
        start = date(2020, 2, 6)
        # If SAT was a non-working day, it would have been the 13th.
        self.assertEqual(
            self.cal.add_working_days(start, 5),
            date(2020, 2, 12)
        )

    def test_no_duplicate_days(self):
        holidays = self.cal.holidays(2021)
        labels = [day[1] for day in holidays]
        labels_dedup = list(set(labels))
        assert sorted(labels) == sorted(labels_dedup)


class HongKongBankTest(HongKongTest):
    cal_class = HongKongBank

    def test_are_saturdays_working_days(self):
        # Let's start with february 6th.
        start = date(2020, 2, 6)
        # If SAT and SUN are non-working days
        self.assertEqual(
            self.cal.add_working_days(start, 5),
            date(2020, 2, 13)
        )


class JapanTest(GenericCalendarTest):

    cal_class = Japan

    def test_year_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)    # new year
        self.assertIn(date(2013, 2, 11), holidays)   # Foundation Day
        self.assertIn(date(2013, 3, 20), holidays)   # Vernal Equinox Day
        self.assertIn(date(2013, 4, 29), holidays)   # Showa Day
        self.assertIn(date(2013, 5, 3), holidays)  # Constitution Memorial Day
        self.assertIn(date(2013, 5, 4), holidays)    # Greenery Day
        self.assertIn(date(2013, 5, 5), holidays)    # Children's Day
        self.assertIn(date(2013, 9, 23), holidays)   # Autumnal Equinox Day
        self.assertIn(date(2013, 11, 3), holidays)   # Culture Day
        self.assertIn(date(2013, 11, 23), holidays)  # Labour Thanksgiving Day
        self.assertIn(date(2013, 12, 23), holidays)  # The Emperor's Birthday

        # Variable days
        self.assertIn(date(2013, 1, 14), holidays)   # Coming of Age Day
        self.assertIn(date(2013, 7, 15), holidays)   # Marine Day
        self.assertIn(date(2013, 9, 16), holidays)   # Respect-for-the-Aged Day
        self.assertIn(date(2013, 10, 14), holidays)  # Health and Sports Day

    def test_year_2016(self):
        # Before 2016, no Mountain Day
        holidays = self.cal.holidays_set(2014)
        self.assertNotIn(date(2014, 8, 11), holidays)   # Mountain Day
        holidays = self.cal.holidays_set(2015)
        self.assertNotIn(date(2015, 8, 11), holidays)   # Mountain Day
        # After 2016, yes
        holidays = self.cal.holidays_set(2016)
        self.assertIn(date(2016, 8, 11), holidays)   # Mountain Day
        holidays = self.cal.holidays_set(2017)
        self.assertIn(date(2017, 8, 11), holidays)   # Mountain Day

    def test_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 1, 14), holidays)  # Coming of Age Day
        self.assertIn(date(2019, 5, 1), holidays)  # Coronation Day
        self.assertIn(date(2019, 5, 6), holidays)  # Children's Day
        self.assertIn(date(2019, 8, 12), holidays)  # Mountain Day Observed
        self.assertIn(date(2019, 11, 4), holidays)  # Culture Day Observed
        # retrieve the sports day label
        holidays = self.cal.holidays(2019)
        holidays_dict = dict(holidays)
        self.assertEqual(
            holidays_dict[date(2019, 10, 14)], "Health and Sports Day")

    def test_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 7, 23), holidays)  # Marine Day
        self.assertIn(date(2020, 7, 24), holidays)  # Sports Day
        self.assertIn(date(2020, 8, 10), holidays)  # Mountain Day adjustment
        self.assertNotIn(date(2020, 8, 11), holidays)  # Mountain Day
        self.assertNotIn(date(2020, 12, 31), holidays)  # New Year's Bank Day
        # retrieve the sports day label
        holidays = self.cal.holidays(2020)
        holidays_dict = dict(holidays)
        self.assertEqual(holidays_dict[date(2020, 7, 24)], "Sports Day")

    def test_year_2021(self):
        holidays = self.cal.holidays_set(2021)
        self.assertIn(date(2021, 7, 22), holidays)  # Marine Day
        self.assertIn(date(2021, 7, 23), holidays)  # Sports Day
        self.assertIn(date(2021, 8, 8), holidays)  # Mountain Day
        self.assertIn(date(2021, 8, 9), holidays)  # Mountain Day Observed
        self.assertNotIn(date(2021, 7, 19), holidays)  # Marine Day
        self.assertNotIn(date(2021, 8, 11), holidays)  # Mountain Day
        self.assertNotIn(date(2021, 10, 11), holidays)  # Sports Day
        # retrieve the sports day label
        holidays = self.cal.holidays(2021)
        holidays_dict = dict(holidays)
        self.assertEqual(holidays_dict[date(2021, 7, 23)], "Sports Day")


class JapanBankTest(GenericCalendarTest):
    cal_class = JapanBank

    def test_year_2019(self):
        holidays = self.cal.holidays_set(2019)
        self.assertIn(date(2019, 1, 2), holidays)  # New Year's Bank Day
        self.assertIn(date(2019, 1, 3), holidays)  # New Year's Bank Day
        self.assertIn(date(2019, 1, 14), holidays)  # Coming of Age Day
        self.assertIn(date(2019, 5, 3), holidays)  # Constitution Memorial Day
        self.assertIn(date(2019, 11, 4), holidays)  # Culture Day Observed
        self.assertIn(date(2019, 12, 31), holidays)  # New Year's Bank Day

    def test_year_2020(self):
        holidays = self.cal.holidays_set(2020)
        self.assertIn(date(2020, 1, 2), holidays)  # New Year's Bank Day
        self.assertIn(date(2020, 1, 3), holidays)  # New Year's Bank Day
        self.assertIn(date(2020, 12, 31), holidays)  # New Year's Bank Day
        self.assertIn(date(2020, 8, 10), holidays)  # Mountain Day adjustment
        self.assertNotIn(date(2020, 8, 11), holidays)  # Mountain Day
        self.assertIn(date(2020, 7, 23), holidays)  # Marine Day
        self.assertIn(date(2020, 7, 24), holidays)  # Sports Day


class MalaysiaTest(GenericCalendarTest):
    cal_class = Malaysia

    def test_year_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)    # New Year's Day
        self.assertIn(date(2013, 1, 28), holidays)   # Thaipusam
        self.assertIn(date(2013, 2, 1), holidays)    # Federal Territory Day
        self.assertIn(date(2013, 2, 11), holidays)   # 2nd day of Lunar NY
        self.assertIn(date(2013, 2, 12), holidays)   # 1st day (Sun lieu)
        self.assertIn(date(2013, 5, 1), holidays)    # Workers' Day
        self.assertIn(date(2013, 5, 24), holidays)   # Vesak Day
        self.assertIn(date(2013, 8, 8), holidays)    # 1st day eid-al-fitr
        self.assertIn(date(2013, 8, 9), holidays)    # 2nd day eid-al-fitr
        self.assertIn(date(2013, 8, 31), holidays)   # National Day
        self.assertIn(date(2013, 9, 16), holidays)   # Malaysia Day
        self.assertIn(date(2013, 10, 15), holidays)  # Hari Raya Haji
        self.assertIn(date(2013, 11, 2), holidays)   # Deepavali
        self.assertIn(date(2013, 11, 5), holidays)   # Islamic New Year
        self.assertIn(date(2013, 12, 25), holidays)  # Xmas

    def test_year_2012(self):
        holidays = self.cal.holidays_set(2012)
        self.assertIn(date(2012, 1, 1), holidays)    # New Year's Day
        self.assertIn(date(2012, 1, 24), holidays)   # Federal Territory Day
        self.assertIn(date(2012, 2, 1), holidays)    # 2nd day of Lunar NY
        self.assertIn(date(2012, 5, 1), holidays)    # 1st day (Sun lieu)
        self.assertIn(date(2012, 5, 5), holidays)    # Workers' Day
        self.assertIn(date(2012, 8, 19), holidays)   # 1st day eid-al-fitr
        self.assertIn(date(2012, 8, 20), holidays)   # 2nd day eid-al-fitr
        self.assertIn(date(2012, 8, 31), holidays)   # National Day
        self.assertIn(date(2012, 9, 16), holidays)   # Malaysia Day
        self.assertIn(date(2012, 10, 26), holidays)  # Hari Raya Haji
        self.assertIn(date(2012, 11, 13), holidays)  # Islamic New Year
        self.assertIn(date(2012, 11, 15), holidays)  # Deepavali
        self.assertIn(date(2012, 12, 25), holidays)  # Xmas

    def test_nuzul_al_quran(self):
        holidays = self.cal.holidays_set(2017)
        self.assertIn(date(2017, 6, 12), holidays)
        holidays = self.cal.holidays_set(2018)
        self.assertIn(date(2018, 6, 1), holidays)

    def test_fix_deepavali_2018(self):
        holidays = self.cal.holidays(2018)
        holidays = dict(holidays)
        deepavali = date(2018, 11, 6)
        self.assertIn(deepavali, holidays)
        self.assertEqual(holidays[deepavali], "Deepavali")

    def test_msia_thaipusam(self):
        # we only have them for years 2010-2024
        self.assertEqual(set(self.cal.MSIA_THAIPUSAM), set(range(2010, 2025)))

    def test_missing_deepavali(self):
        save_2020 = self.cal.MSIA_DEEPAVALI[2020]
        del self.cal.MSIA_DEEPAVALI[2020]
        with self.assertRaises(KeyError):
            self.cal.holidays(2020)
        # Back to normal, to avoid breaking further tests
        self.cal.MSIA_DEEPAVALI[2020] = save_2020

    def test_missing_thaipusam(self):
        save_2020 = self.cal.MSIA_THAIPUSAM[2020]
        del self.cal.MSIA_THAIPUSAM[2020]
        with self.assertRaises(KeyError):
            self.cal.holidays(2020)
        # Back to normal, to avoid breaking further tests
        self.cal.MSIA_THAIPUSAM[2020] = save_2020

    def test_labour_day_label(self):
        holidays = self.cal.holidays(2020)
        holidays = dict(holidays)
        self.assertEqual(
            holidays[date(2020, 5, 1)], "Workers' Day")


class QatarTest(GenericCalendarTest):
    cal_class = Qatar
    test_include_january_1st = False

    def test_year_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 7, 9), holidays)  # start ramadan
        # warning, the official date was (2013, 8, 10)
        self.assertIn(date(2013, 8, 8), holidays)  # eid al fitr
        # The official date was (2013, 10, 14)
        self.assertIn(date(2013, 10, 15), holidays)  # eid al adha
        self.assertIn(date(2013, 10, 16), holidays)  # eid al adha
        self.assertIn(date(2013, 10, 17), holidays)  # eid al adha
        self.assertIn(date(2013, 10, 18), holidays)  # eid al adha
        self.assertIn(date(2013, 12, 18), holidays)  # National Day

    def test_weekend(self):
        # In Qatar, Week-end days are Friday / Sunday.
        weekend_day = date(2017, 5, 12)  # This is a Friday
        non_weekend_day = date(2017, 5, 14)  # This is a Sunday
        self.assertFalse(self.cal.is_working_day(weekend_day))
        self.assertTrue(self.cal.is_working_day(non_weekend_day))


class SingaporeTest(GenericCalendarTest):

    cal_class = Singapore

    def test_CNY_2010(self):
        holidays = self.cal.holidays_set(2010)
        self.assertIn(date(2010, 2, 14), holidays)  # CNY1
        self.assertIn(date(2010, 2, 15), holidays)  # CNY2
        self.assertIn(date(2010, 2, 16), holidays)  # Rolled day for CNY

    def test_year_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)  # New Year
        self.assertIn(date(2013, 2, 10), holidays)  # CNY1
        self.assertIn(date(2013, 2, 11), holidays)  # CNY2
        self.assertIn(date(2013, 2, 12), holidays)  # Rolled day for CNY
        self.assertIn(date(2013, 3, 29), holidays)  # Good Friday
        self.assertIn(date(2013, 5, 1), holidays)  # Labour Day
        self.assertIn(date(2013, 5, 24), holidays)  # Vesak Day
        self.assertIn(date(2013, 8, 8), holidays)  # Hari Raya Puasa
        self.assertIn(date(2013, 8, 9), holidays)  # National Day
        self.assertIn(date(2013, 10, 15), holidays)  # Hari Raya Haji
        self.assertIn(date(2013, 11, 3), holidays)  # Deepavali
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2013, 11, 4), observed)  # Deepavali shift
        self.assertIn(date(2013, 12, 25), holidays)  # Christmas Day

    def test_year_2018(self):
        holidays = self.cal.holidays_set(2018)
        self.assertIn(date(2018, 1, 1), holidays)  # New Year
        self.assertIn(date(2018, 2, 16), holidays)  # CNY
        self.assertIn(date(2018, 2, 17), holidays)  # CNY
        self.assertIn(date(2018, 3, 30), holidays)  # Good Friday
        self.assertIn(date(2018, 5, 1), holidays)  # Labour Day
        self.assertIn(date(2018, 5, 29), holidays)  # Vesak Day
        self.assertIn(date(2018, 6, 15), holidays)  # Hari Raya Puasa
        self.assertIn(date(2018, 8, 9), holidays)  # National Day
        self.assertIn(date(2018, 8, 22), holidays)  # Hari Raya Haji
        self.assertIn(date(2018, 11, 6), holidays)  # Deepavali
        self.assertIn(date(2018, 12, 25), holidays)  # Christmas Day

    def test_fixed_holiday_shift(self):
        # Labour Day was on a Sunday in 2016
        holidays = self.cal.holidays_set(2016)
        # Labour Day (sunday)
        self.assertIn(date(2016, 5, 1), holidays)
        # Shifted day (Monday)
        observed = set(map(self.cal.get_observed_date, holidays))
        self.assertIn(date(2016, 5, 2), observed)

    def test_deepavali(self):
        # At the moment, we have values for deepavali only until year 2021
        for year in range(2000, 2022):
            self.assertIn(year, self.cal.DEEPAVALI)

    def test_deepavali_current_year(self):
        # This is a regression test. We should be able to have the deepavali
        # value until this year and probably a few years in the future
        current_year = date.today().year
        self.assertIn(current_year, self.cal.DEEPAVALI)

    def test_deepavali_missing_year(self):
        with self.assertRaises(KeyError) as context:
            self.cal.holidays_set(1999)
        self.assertEqual(
            # Equivalent of the error msg
            context.exception.args[0],
            'Missing date for Singapore Deepavali for year: 1999',
        )


class SouthKoreaTest(GenericCalendarTest):

    cal_class = SouthKorea

    def test_year_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)    # new year
        self.assertIn(date(2013, 3, 1), holidays)    # Independence day
        self.assertIn(date(2013, 5, 5), holidays)    # children's day
        self.assertIn(date(2013, 6, 6), holidays)    # Memorial day
        self.assertIn(date(2013, 8, 15), holidays)   # Liberation day
        self.assertIn(date(2013, 10, 3), holidays)   # National Foundation Day
        self.assertIn(date(2013, 10, 9), holidays)   # Hangul Day
        self.assertIn(date(2013, 12, 25), holidays)  # Christmas

        # Variable days
        self.assertIn(date(2013, 2, 9), holidays)
        self.assertIn(date(2013, 2, 10), holidays)
        self.assertIn(date(2013, 2, 11), holidays)
        self.assertIn(date(2013, 5, 17), holidays)
        self.assertIn(date(2013, 9, 18), holidays)
        self.assertIn(date(2013, 9, 19), holidays)
        self.assertIn(date(2013, 9, 20), holidays)


class TaiwanTest(GenericCalendarTest):

    cal_class = Taiwan

    def test_year_2013(self):
        holidays = self.cal.holidays_set(2013)
        self.assertIn(date(2013, 1, 1), holidays)    # New Year
        self.assertIn(date(2013, 2, 9), holidays)    # Chinese new year's eve
        self.assertIn(date(2013, 2, 10), holidays)   # Chinese new year
        self.assertIn(date(2013, 2, 11), holidays)   # Spring Festival
        self.assertIn(date(2013, 2, 12), holidays)   # Spring Festival
        self.assertIn(date(2013, 2, 28), holidays)   # 228 Peace Memorial Day
        self.assertIn(date(2013, 4, 4), holidays)    # Children's Day
        self.assertIn(date(2013, 6, 12), holidays)   # Dragon Boat Festival
        self.assertIn(date(2013, 9, 19), holidays)   # Mid-Autumn Festival
        self.assertIn(date(2013, 10, 10), holidays)  # National Day

    def test_qingming_festival(self):
        self.assertIn(date(2001, 4, 5), self.cal.holidays_set(2001))
        self.assertIn(date(2002, 4, 5), self.cal.holidays_set(2002))
        self.assertIn(date(2005, 4, 5), self.cal.holidays_set(2005))
        self.assertIn(date(2006, 4, 5), self.cal.holidays_set(2006))
        self.assertIn(date(2007, 4, 5), self.cal.holidays_set(2007))
        self.assertIn(date(2008, 4, 4), self.cal.holidays_set(2008))
        self.assertIn(date(2010, 4, 5), self.cal.holidays_set(2010))
        self.assertIn(date(2011, 4, 5), self.cal.holidays_set(2011))
        self.assertIn(date(2012, 4, 4), self.cal.holidays_set(2012))
        self.assertIn(date(2013, 4, 4), self.cal.holidays_set(2013))
        self.assertIn(date(2014, 4, 4), self.cal.holidays_set(2014))

    def test_extra_day_2021_february(self):
        extra_day_feb_20 = date(2021, 2, 20)
        self.assertTrue(self.cal.is_working_day(extra_day_feb_20))
        # Compute deltas with the extra-working day...
        five_days_after = self.cal.add_working_days(
            date(2021, 2, 19), 5
        )
        # Should've been on 26th, albeit the 20th
        self.assertEqual(five_days_after, date(2021, 2, 25))

    def test_extra_day_2021_september(self):
        extra_day_sept_11 = date(2021, 9, 11)
        self.assertTrue(self.cal.is_working_day(extra_day_sept_11))
        # Compute deltas with the extra-working day...
        five_days_after = self.cal.add_working_days(
            date(2021, 9, 10), 5
        )
        # Should've been on 17th, albeit the 11th
        self.assertEqual(five_days_after, date(2021, 9, 16))


class IsraelTest(GenericCalendarTest):

    cal_class = Israel
    test_include_january_1st = False

    def test_holidays_2017(self):
        calculated_holidays = self.cal.holidays_set(2017)
        known_holidays = {
            date(2017, 4, 10),  # Passover (Pesach)
            date(2017, 4, 11),
            date(2017, 4, 16),
            date(2017, 4, 17),
            date(2017, 5, 1),  # Independence Day (Yom Ha-Atzmaut)
            date(2017, 5, 2),
            date(2017, 9, 20),  # Jewish New Year (Rosh Ha-Shana)
            date(2017, 9, 21),
            date(2017, 9, 22),
            date(2017, 9, 29),  # Yom Kippur
            date(2017, 9, 30),
            date(2017, 10, 4),  # Sukkot
            date(2017, 10, 5),
            date(2017, 10, 11),
            date(2017, 10, 12),
            date(2017, 5, 30),  # Shavuot
            date(2017, 5, 31),
        }
        self.assertEqual(calculated_holidays, known_holidays)

    def test_holidays_2018(self):
        calculated_holidays = self.cal.holidays_set(2018)
        known_holidays = {
            date(2018, 3, 30),  # Passover (Pesach)
            date(2018, 3, 31),
            date(2018, 4, 5),
            date(2018, 4, 6),
            date(2018, 4, 18),  # Independence Day (Yom Ha-Atzmaut)
            date(2018, 4, 19),
            date(2018, 9, 9),  # Rosh Hashana
            date(2018, 9, 10),
            date(2018, 9, 11),
            date(2018, 9, 18),  # Yom Kippur
            date(2018, 9, 19),
            date(2018, 9, 23),  # Sukkot
            date(2018, 9, 24),
            date(2018, 9, 30),
            date(2018, 10, 1),
            date(2018, 5, 19),  # Shavuot
            date(2018, 5, 20),
        }
        self.assertEqual(calculated_holidays, known_holidays)

    def test_holidays_2019(self):
        calculated_holidays = self.cal.holidays_set(2019)
        known_holidays = {
            date(2019, 4, 19),  # Passover (Pesach)
            date(2019, 4, 20),  # Passover (Pesach)
            date(2019, 4, 25),  # Passover (Pesach)
            date(2019, 4, 26),  # Passover (Pesach)
            date(2019, 5, 8),  # Independence Day (Yom Ha-Atzmaut)
            date(2019, 5, 9),  # Independence Day (Yom Ha-Atzmaut)
            date(2019, 6, 8),  # Shavuot
            date(2019, 6, 9),  # Shavuot
            date(2019, 9, 29),  # Rosh Hashana
            date(2019, 9, 30),  # Rosh Hashana
            date(2019, 10, 1),  # Rosh Hashana
            date(2019, 10, 8),  # Yom Kippur
            date(2019, 10, 9),  # Yom Kippur
            date(2019, 10, 13),  # Sukkot
            date(2019, 10, 14),  # Sukkot
            date(2019, 10, 20),  # Sukkot
            date(2019, 10, 21),  # Sukkot
        }
        self.assertEqual(calculated_holidays, known_holidays)

    def test_holidays_2020(self):
        calculated_holidays = self.cal.holidays_set(2020)
        known_holidays = {
            date(2020, 9, 18),  # Rosh Hashana Eve
            date(2020, 9, 19),  # Rosh Hashana A
            date(2020, 9, 20),  # Rosh Hashana B
            date(2020, 9, 27),  # Kippur Eve
            date(2020, 9, 28),  # Kippur
            date(2020, 10, 2),  # Sukot A Eve
            date(2020, 10, 3),  # Sukot A
            date(2020, 10, 9),  # Sukot B Eve
            date(2020, 10, 10),  # Sukot B
            date(2020, 4, 8),  # Pesach A Eve
            date(2020, 4, 9),  # Pesach A
            date(2020, 4, 14),  # Pesach B Eve
            date(2020, 4, 15),  # Pesach B
            date(2020, 4, 28),  # Independence Day Eve
            date(2020, 4, 29),  # Independence Day
            date(2020, 5, 28),  # Shavuot Eve
            date(2020, 5, 29),  # Shavuot
        }
        self.assertEqual(calculated_holidays, known_holidays)

    def test_holidays_2021(self):
        calculated_holidays = self.cal.holidays_set(2021)
        known_holidays = {
            date(2021, 9, 6),  # Rosh Hashana Eve
            date(2021, 9, 7),  # Rosh Hashana A
            date(2021, 9, 8),  # Rosh Hashana B
            date(2021, 9, 15),  # Kippur Eve
            date(2021, 9, 16),  # Kippur
            date(2021, 9, 20),  # Sukot A Eve
            date(2021, 9, 21),  # Sukot A
            date(2021, 9, 27),  # Sukot B Eve
            date(2021, 9, 28),  # Sukot B
            date(2021, 3, 27),  # Pesach A Eve
            date(2021, 3, 28),  # Pesach A
            date(2021, 4, 2),  # Pesach B Eve
            date(2021, 4, 3),  # Pesach B
            date(2021, 4, 14),  # Independence Day Eve
            date(2021, 4, 15),  # Independence Day
            date(2021, 5, 16),  # Shavuot Eve
            date(2021, 5, 17),  # Shavuot
        }
        self.assertEqual(calculated_holidays, known_holidays)

    def test_is_holiday_performance(self):
        random_date = date(2019, 10, 9)
        japan_cal = Japan()
        timer = time.time()
        for i in range(30):
            japan_cal.is_holiday(random_date)
        japan_time = time.time() - timer
        timer = time.time()
        for i in range(30):
            self.cal.is_holiday(random_date)
        israel_time = time.time() - timer
        self.assertGreater(japan_time * 3, israel_time)



from datetime import date, timedelta
from ..core import WesternCalendar, SUN
from ..registry_tools import iso_register


@iso_register('CH')
class Switzerland(WesternCalendar):
    'Switzerland'

    # ChristianMixin entries common to (most) cantons - opt out
    include_good_friday = True  # not in TI, VS
    include_easter_sunday = True
    include_easter_monday = True  # not in VS
    include_ascension = True
    include_whit_sunday = True
    include_whit_monday = True  # not in VS
    include_christmas = True
    include_boxing_day = True  # not in GE, NE, VS, VD

    # ChristianMixin entries with varying observance - opt in
    include_epiphany = False
    include_corpus_christi = False
    include_assumption = False
    include_all_saints = False
    include_immaculate_conception = False

    # Swiss entries with varying observance - opt in
    include_berchtolds_day = False
    include_st_josephs_day = False

    FIXED_HOLIDAYS = WesternCalendar.FIXED_HOLIDAYS + (
        (8, 1, "National Holiday"),
    )

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        if self.include_berchtolds_day:
            days.append((date(year, 1, 2), "Berchtold's Day"))
        if self.include_st_josephs_day:
            days.append((date(year, 3, 19), "St Joseph's Day"))
        return days


@iso_register('CH-AG')
class Aargau(Switzerland):
    'Aargau'

    include_berchtolds_day = True
    include_corpus_christi = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-AI')
class AppenzellInnerrhoden(Switzerland):
    'Appenzell Innerrhoden'

    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-AR')
class AppenzellAusserrhoden(Switzerland):
    'Appenzell Ausserrhoden'

    include_labour_day = True


@iso_register('CH-BE')
class Bern(Switzerland):
    'Bern'

    include_berchtolds_day = True


@iso_register('CH-BL')
class BaselLandschaft(Switzerland):
    'Basel-Landschaft'

    include_labour_day = True


@iso_register('CH-BS')
class BaselStadt(Switzerland):
    'Basel-Stadt'

    include_labour_day = True


@iso_register('CH-FR')
class Fribourg(Switzerland):
    'Fribourg'

    include_berchtolds_day = True
    include_labour_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-GE')
class Geneva(Switzerland):
    'Geneva'

    include_boxing_day = False

    FIXED_HOLIDAYS = Switzerland.FIXED_HOLIDAYS + (
        (12, 31, "Creation of Geneva Republic"),
    )

    def get_genevan_fast(self, year):
        "Thursday following the first Sunday of September"
        first_sunday = self.get_nth_weekday_in_month(year, 9, SUN)
        # The following thursday is 4 days after
        return (
            first_sunday + timedelta(days=4),
            "Genevan Fast"
        )

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(self.get_genevan_fast(year))
        return days


@iso_register('CH-GL')
class Glarus(Switzerland):
    'Glarus (Glaris)'

    include_berchtolds_day = True
    include_all_saints = True

    FIXED_HOLIDAYS = Switzerland.FIXED_HOLIDAYS + (
        (4, 3, "Näfels Ride"),
    )


@iso_register('CH-GR')
class Graubunden(Switzerland):
    'Graubünden (Grisons)'

    include_epiphany = True
    include_st_josephs_day = True
    include_corpus_christi = True
    include_immaculate_conception = True


@iso_register('CH-JU')
class Jura(Switzerland):
    'Jura'

    include_berchtolds_day = True
    include_labour_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_boxing_day = False

    FIXED_HOLIDAYS = Switzerland.FIXED_HOLIDAYS + (
        (6, 23, "Independence Day"),
    )


@iso_register('CH-LU')
class Luzern(Switzerland):
    'Luzern'

    include_berchtolds_day = True
    include_epiphany = True
    include_st_josephs_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-NE')
class Neuchatel(Switzerland):
    'Neuchâtel'

    include_boxing_day = False
    include_berchtolds_day = True
    include_labour_day = True

    FIXED_HOLIDAYS = Switzerland.FIXED_HOLIDAYS + (
        (3, 1, "Republic Day"),
    )


@iso_register('CH-NW')
class Nidwalden(Switzerland):
    'Nidwalden'

    include_st_josephs_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-OW')
class Obwalden(Switzerland):
    'Obwalden'

    include_berchtolds_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True

    FIXED_HOLIDAYS = Switzerland.FIXED_HOLIDAYS + (
        (9, 25, "Saint Nicholas of Flüe Day"),
    )


@iso_register('CH-SG')
class StGallen(Switzerland):
    'St. Gallen'

    include_all_saints = True


@iso_register('CH-SH')
class Schaffhausen(Switzerland):
    'Schaffhausen'

    include_berchtolds_day = True
    include_labour_day = True


@iso_register('CH-SO')
class Solothurn(Switzerland):
    'Solothurn'

    include_berchtolds_day = True
    include_st_josephs_day = True
    include_labour_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-SZ')
class Schwyz(Switzerland):
    'Schwyz'

    include_epiphany = True
    include_st_josephs_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-TG')
class Thurgau(Switzerland):
    'Thurgau'

    include_berchtolds_day = True
    include_labour_day = True


@iso_register('CH-TI')
class Ticino(Switzerland):
    'Ticino'

    include_good_friday = False
    include_epiphany = True
    include_st_josephs_day = True
    include_labour_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True

    FIXED_HOLIDAYS = Switzerland.FIXED_HOLIDAYS + (
        (6, 29, "Saints Peter and Paul"),
    )


@iso_register('CH-UR')
class Uri(Switzerland):
    'Uri'

    include_epiphany = True
    include_st_josephs_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-VD')
class Vaud(Switzerland):
    'Vaud'

    include_berchtolds_day = True
    include_boxing_day = False

    def get_federal_thanksgiving_monday(self, year):
        "Monday following the 3rd sunday of September"
        third_sunday = self.get_nth_weekday_in_month(year, 9, SUN, 3)
        return (
            third_sunday + timedelta(days=1),
            "Federal Thanksgiving Monday"
        )

    def get_variable_days(self, year):
        days = super().get_variable_days(year)
        days.append(self.get_federal_thanksgiving_monday(year))
        return days


@iso_register('CH-VS')
class Valais(Switzerland):
    'Valais'

    include_good_friday = False
    include_easter_monday = False
    include_whit_monday = False
    include_st_josephs_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True
    include_boxing_day = False


@iso_register('CH-ZG')
class Zug(Switzerland):
    'Zug'

    include_berchtolds_day = True
    include_corpus_christi = True
    include_assumption = True
    include_all_saints = True
    include_immaculate_conception = True


@iso_register('CH-ZH')
class Zurich(Switzerland):
    'Zürich'

    include_berchtolds_day = True
    include_labour_day = True
